#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import logging
import re
import six
import sys
import traceback
import asyncio
import sqlalchemy
from sqlalchemy.orm import query, loading, attributes
from sqlalchemy.ext.asyncio import AsyncSession, AsyncEngine, AsyncConnection, create_async_engine
import sqlalchemy.databases as sqlalchemy_supported_engines
# from sqlalchemy.orm import sessionmaker
import motor.motor_asyncio
import motor
import sqlalchemy.sql.dml
from sqlalchemy.sql.expression import delete, insert
from sshtunnel import SSHTunnelForwarder

# from tornado.ioloop import IOLoop
# target_module = __import__('tornado.concurrent')
# if not hasattr(target_module.concurrent, 'return_future'):
#     setattr(target_module.concurrent, 'return_future', tornado.gen.coroutine)
# from motorengine import connect as mongo_connect
import mongoengine
import pymongo

from .sqlalchemy_dialects import asyncpg_migrate
from .sqlalchemy_dialects.sync_threading import AsyncioEngine as ThreadingAsyncioEngine

try:
    import asyncpg
except ImportError:
    pass
try:
    import aiomysql
except ImportError:
    pass
try:
    import aiosqlite
except ImportError:
    pass

try:
    import pymssql
except ImportError:
    pass
try:
    import cx_Oracle
    # import cx_Oracle_async
except ImportError:
    cx_Oracle = {}

from .supports import singleton
from .modelutils import model_columns, format_mongo_value, get_dbinstance_by_model, get_model_class_name, DEFAULT_SKIP_FIELDS
from .utilities import url_decode, url_encode
from .exceptionreporter import ExceptionReporter

logging.getLogger('aiosqlite').setLevel(logging.INFO)
LOG = logging.getLogger('lycium.dbproxy')

supported_engines = [*sqlalchemy_supported_engines.__all__, 'cockroachdb']
supported_asyncio_engines = ['postgresql', 'mysql', 'cockroachdb', 'sqlite']

class _DbInstance(object):
    """
    """
    def __init__(self, name: str, engine: AsyncEngine, should_connect: bool = False, connection_description: str = '', ssh_tunnel: SSHTunnelForwarder = None) -> None:
        self.name: str = name
        self.engine: AsyncEngine = engine
        self.connection_description = connection_description
        self.ssh_tunnel = ssh_tunnel
        self.isconnecting: bool = False
        self.disconnected: bool = True
        self.conn: AsyncConnection = None
        if should_connect:
            self._start_connect()
        self.async_by_thread = False
        if hasattr(engine, 'async_by_thread'):
            self.async_by_thread = getattr(engine, 'async_by_thread')

    def onconnected(self, conn: AsyncConnection) -> None:
        self.disconnected = False
        self.isconnecting = False
        self.conn = conn

    def ondisconnected(self, reason: str, reconnect: bool = False) -> None:
        self.disconnected = True
        self.isconnecting = False
        if self.conn:
            self.conn.close()
            self.conn = None
        LOG.warn('db connection %s disconnected with reason:%s', self.name, reason)
        if reconnect:
            reconnect_secs = 2
            LOG.info('db connection %s will be reconnecting after %d seconds', self.name, reconnect_secs)
            asyncio.get_event_loop().call_later(reconnect_secs, self._connect_db)
            # IOLoop.current().call_later(reconnect_secs, self._connect_db)

    async def manual_connect(self):
        return await self._connect_db()

    async def _connect_db(self):
        if self.isconnecting:
            return
        connect_desc = self.get_connection_description()
        try:
            LOG.info('connecting database %s %s', self.name, connect_desc)
            t1 = time.time()
            self.isconnecting = True
            db_conn = await self.engine.connect()
            t2 = time.time()
            LOG.info('database %s connected in %.2f secs.', self.name, t2 - t1)
            self.onconnected(db_conn)
        except Exception as e:
            ExceptionReporter().report(key='DB-'+connect_desc, typ='CONNECT', 
                endpoint='%s|%s' % ('CONNECT', connect_desc),
                method='CONNECT',
                inputs=str(self.engine),
                outputs=str(e),
                content=traceback.format_exc(),
                level='ERROR'
            )
            self.ondisconnected(str(e), True)
    
    def _start_connect(self):
        try:
            loop = asyncio.get_event_loop()
        except Exception as e:
            LOG.info('get current asyncio event loop failed with error:%s, creates new event loop')
            loop = asyncio.new_event_loop()
        
        LOG.info('registering future %s connecting for %s in asyncio', self.engine.name, self.name)
        asyncio.run_coroutine_threadsafe(self._connect_db(), loop)

    def begin(self):
        return self.engine.begin()

    def get_connection_description(self) -> str:
        if self.connection_description:
            return self.connection_description
        
        url_info = self.engine.url
        auth_part = ''
        host_part = ''
        if url_info.username or url_info.password:
            auth_part = '%s:%s@' % ((url_info.username if url_info.username else ''), ('*' * len(url_info.password) if url_info.password else ''))
        if url_info.host:
            host_part = '%s:%d' % (url_info.host, url_info.port)
        self.connection_description = '%s://%s%s/%s' % (url_info.drivername, auth_part, host_part, url_info.database)
        if url_info.normalized_query:
            url_args = []
            for k, v in url_info.normalized_query.items():
                if len(v) > 0:
                    url_args.append(str(k)+'='+','.join(v))
            self.connection_description = self.connection_description + '?' + '&'.join(url_args)

@singleton
class DbProxy(object):
    """
    database agent component
    """
    def __init__(self):
        self.db_instances = {}
        self.default_mongo_db_instance = None
        self.default_rdbms_db_instance = None
        self._motor_count_documents_name = 'count_documents'

        if hasattr(motor.MotorCollection, 'count'):
            self._motor_count_documents_name = 'count'

        self._cur_execution_dbinstances = []

    def setup_rdbms(self, rdbms_configs: dict) -> bool:
        """Setup relational database configurations
        :param rdbms_configs:dict relational database configuration, it would skip the configuration named start with 'ignore'
            examples: {
                "local_storage": {
                    "connector": "sqlite",
                    "host": "./debug.sqlite.db",
                    "db": "main"
                },
                "remote_relational": {
                    "connector": "mysql",
                    "host": "127.0.0.1",
                    "port": 3306,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "changeit"
                },
                "remote_relational_oracle_connect_with_sid": {
                    "connector": "oracle",
                    "host": "127.0.0.1",
                    "port": 1521,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "orcl"
                },
                "remote_relational_oracle_connect_with_service_name": {
                    "connector": "oracle",
                    "host": "127.0.0.1",
                    "port": 1521,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "",
                    "connect_args": {"encoding": "UTF-16", "nencoding": "UTF-16"},
                    "ext_args": {"service_name": "orcl"}
                },
                "ignore_relational": {      # the category name starts with 'ignore' characters would skip connecting
                    "connector": "mssql",
                    "host": "127.0.0.1",
                    "port": 1433,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "changeit"
                }
           }
        :return: bool
        """
        for k, dbconf in rdbms_configs.items():
            self.setup_rdbms_connection(k, dbconf)

    def setup_mongodbs(self, mongodb_configs: dict) -> bool:
        """Setup mongodb database configurations
        :param rdbms_configs:dict mongodb database configuration, it would skip the configuration named start with 'ignore'
            examples: {
                "ignore_mongo": {           # the category name starts with 'ignore' characters would skip connecting
                    "connector": "mongodb",
                    "host":  "127.0.0.1",
                    "port": 27017,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "changeit"
                }
            }
        :return: bool
        """
        for k, dbconf in mongodb_configs.items():
            self.setup_mongodb_connection(k, dbconf)

    def setup_rdbms_connection(self, category: str, dbconf: dict) -> bool:
        """Setup relational database configuration by configuration, the database instance would be named by category
        :param category:str name of rdbms configuration dict key
        :param dbconf:dict value of rdbms configuration element, contains keys of connector, host, port, user, pwd, db
            example: {
                    "connector": "oracle",
                    "host": "127.0.0.1",
                    "port": 1521,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "changeit"
                }
        :return: bool
        """
        if category.startswith('ignore'):
            return False

        engine = dbconf.get('connector', '')
        if engine not in supported_engines:
            LOG.error('setup rdbms connection with engine name:%s were not supported, skip it', engine)
            return False

        conndsn, connect_description, ssh_tunnel = self.format_connection_string(dbconf)
        create_engine_params = self.format_create_engine_parameters(engine=engine, dbconf=dbconf)
        LOG.info('initializing database engine for %s', category)
        if engine in supported_asyncio_engines:
            db_engine = create_async_engine(conndsn, **create_engine_params)
        else:
            sync_engine = sqlalchemy.create_engine(conndsn, **create_engine_params)
            db_engine = ThreadingAsyncioEngine(sync_engine)
        db_inst = _DbInstance(category, db_engine, should_connect=True, connection_description=connect_description, ssh_tunnel=ssh_tunnel)
        self.db_instances[category] = db_inst
        if not self.default_rdbms_db_instance:
            self.default_rdbms_db_instance = db_inst

        return True

    def setup_mongodb_connection(self, category: str, dbconf: dict) -> bool:
        """Setup mongodb database configuration by configuration, the database instance would be named by category
        :param category:str name of mongodb configuration dict key
        :param dbconf:dict value of mongodb configuration element, contains keys of connector, host, port, user, pwd, db
            example: {
                    "connector": "mongodb",
                    "host":  "127.0.0.1",
                    "port": 27017,
                    "user": "changeit",
                    "pwd": "changeit",      # password
                    "db": "changeit"
                }
        :return: bool
        """
        if category.startswith('ignore'):
            return False
        conndsn, connect_description, ssh_tunnel = self.format_connection_string(dbconf)
        mongo_cli = motor.motor_asyncio.AsyncIOMotorClient(conndsn)
        mongo_db = mongo_cli[dbconf.get('db')]
        LOG.info('mongodb %s %s setupped', category, str(mongo_cli.address))
        db_inst = _DbInstance(category, mongo_cli, should_connect=False, connection_description=connect_description, ssh_tunnel=ssh_tunnel)
        db_inst.disconnected = True
        db_inst.conn = mongo_db
        if not self.default_mongo_db_instance:
            self.default_mongo_db_instance = db_inst
        self.db_instances[category] = db_inst
        return True

    def format_create_engine_parameters(self, engine: str, dbconf: dict) -> dict:
        create_engine_params = {
            'encoding': dbconf.get('encoding', 'utf8'),
            'pool_pre_ping': True,
            'pool_use_lifo': True,
            'pool_timeout': 10,
            'pool_size': 5,
            'pool_recycle': 3600,
            'echo_pool': True,
            'max_overflow': 10,
            # 'echo': True,
            # 'strategy': ASYNCIO_STRATEGY,
            'connect_args': {
                # 'connect_timeout': 15,
                # 'timeout': 15
            }
        }
        if 'sqlite' == engine:
            for unsupported_key in ['pool_timeout', 'pool_size', 'pool_use_lifo', 'max_overflow']:
                if unsupported_key in create_engine_params:
                    del create_engine_params[unsupported_key]
        elif 'mssql' == engine:
            create_engine_params['isolation_level'] = 'AUTOCOMMIT'
        if 'connect_args' in dbconf:
            connect_args = dbconf['connect_args']
            for custom_key in ['pool_size', 'pool_timeout', 'pool_recycle', 'max_overflow', 'isolation_level', 'pool_use_lifo']:
                if custom_key in connect_args:
                    if custom_key in ['isolation_level']:
                        create_engine_params[custom_key] = str(connect_args[custom_key])
                    elif custom_key in ['pool_use_lifo']:
                        create_engine_params[custom_key] = bool(connect_args[custom_key])
                    else:
                        create_engine_params[custom_key] = int(connect_args[custom_key])
                    del connect_args[custom_key]
                    dbconf['connect_args'] = connect_args
            if not isinstance(connect_args, dict):
                s1 = str(connect_args).split('&')
                connect_args = {}
                for s2 in s1:
                    s3 = s2.split('=')
                    if len(s3) > 1:
                        connect_args[s3[0].strip()] = s3[1].strip()
            for k, v in connect_args.items():
                create_engine_params['connect_args'][k] = v
        return create_engine_params

    def format_connection_string(self, dbconf: dict) -> str:
        """Formats database connection sqlalchemy formation dsn string
        :param dbconf:dict value of mongodb configuration element, contains keys of connector, host, port, user, pwd, db
        :return:str example: postgresql+asyncpg://user:password@host:port/db
        """
        engine = dbconf.get('connector', 'mysql')
        user = dbconf.get('user', 'guest')
        pwd = dbconf.get('pwd', '')
        host = dbconf.get('host', 'localhost')
        port = int(dbconf.get('port', 3306))
        dbname = dbconf.get('db', 'guest')
        driverpart = dbconf.get('driver', '')
        if driverpart:
            driverpart = '+' + driverpart
        ext_params = ''
        if dbconf.get('ext_args'):
            ext_args = [f'{ext_arg_name}={ext_arg_value}' for ext_arg_name, ext_arg_value in
                        dbconf.get('ext_args', {}).items()]
            ext_params = f'?{"&".join(ext_args)}'
        if 'cockroachdb' == engine:
            engine = 'postgresql'
            driverpart = '+asyncpg'
        elif 'postgresql' == engine:
            driverpart = '+asyncpg'
        elif 'mysql' == engine:
            driverpart = '+aiomysql'
        elif 'sqlite' == engine:
            driverpart = '+aiosqlite'
        elif 'oracle' == engine:
            driverpart = '+cx_oracle'
        elif 'mssql' == engine:
            driverpart = '+pymssql'

        connection_str = ''
        connection_description = ''
        ssh_tunnel = None
        password_description = ('*' * len(pwd) if pwd else '')

        connect_host = host
        connect_port = port
        if 'sshtunnel' in dbconf:
            connect_host, connect_port, ssh_tunnel = self.make_ssh_tunnel(dbconf['sshtunnel'], host, port, engine, dbconf)
        if 'oracle' == engine:
            if 'service_name' in dbconf.get('ext_args', {}):
                if '' != dbname:
                    dbname = ''
            # connection_str = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), url_encode(pwd), connect_host, connect_port, dbname, ext_params)
            # connection_description = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), password_description, host, port, dbname, ext_params)
            # ora_dsn = cx_Oracle.makedsn(host, port, service_name=dbname)
            # connection_str = "%s%s://%s:%s@%s%s" % (engine, driverpart, user, url_encode(pwd), ora_dsn, ext_params)
        if 'sqlite' == engine:
            connection_str = '%s%s:///%s' % (engine, driverpart, host)
            connection_description = connection_str
        elif 'mongodb' == engine:
            if dbconf.get('url'):
                connection_str = dbconf.get('url')
                connection_description = connection_str
            else:
                if connect_port == 0:
                    connection_str = '%s%s://%s:%s@%s' % (engine, driverpart, url_encode(user), url_encode(pwd), connect_host)
                else:
                    connection_str = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), url_encode(pwd), connect_host, connect_port, dbname, ext_params)
                connection_description = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), password_description, host, port, dbname, ext_params)
        else:
            if connect_port == 0:
                connection_str = '%s%s://%s:%s@%s' % (engine, driverpart, url_encode(user), url_encode(pwd), connect_host)
            else:
                connection_str = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), url_encode(pwd), connect_host, connect_port, dbname, ext_params)
            connection_description = '%s%s://%s:%s@%s:%d/%s%s' % (engine, driverpart, url_encode(user), password_description, host, port, dbname, ext_params)
        return connection_str, connection_description, ssh_tunnel

    def parse_ssh_address(self, ssh_address: str):
        ssh_host = ssh_address
        ssh_user = ''
        ssh_password = ''
        conn_parts = ssh_address.split('://')
        if len(conn_parts) > 1:
            ssh_address = conn_parts[1]
        else:
            ssh_address = conn_parts[0]
        auth_parts = ssh_address.split('@')
        if len(auth_parts) > 1:
            auth_texts = auth_parts[0].split(':')
            ssh_user = url_decode(auth_texts[0])
            if len(auth_texts) > 1:
                ssh_password = url_decode(auth_texts[1])
            ssh_host = auth_parts[1]
        else:
            ssh_host = auth_parts[0]
        host_parts = ssh_host.split(':')
        ssh_host = (host_parts[0], int(host_parts[1] if len(host_parts) > 1 else 22))
        return ssh_user, ssh_password, ssh_host

    def make_ssh_tunnel(self, ssh_address: str, remote_host: str, remote_port: int, engine: str, dbconf: dict):
        remote_addresses = [(remote_host, remote_port)]
        if 'replica_hosts' in dbconf:
            for h in dbconf['replica_hosts']:
                hs = h.split(':')
                hport = int(hs[1]) if len(hs) > 1 else self.get_default_db_port(dbconf.get('connector', 'mysql'))
                remote_addresses.append((hs[0], hport))
        ssh_user, ssh_password, ssh_host = self.parse_ssh_address(ssh_address)
        ssh_tunnel = SSHTunnelForwarder(ssh_host,
                                        ssh_username=ssh_user,
                                        ssh_password=ssh_password,
                                        remote_bind_addresses=remote_addresses,
                                        threaded=True)
        ssh_tunnel.start()
        local_addresses = ssh_tunnel.local_bind_addresses
        connect_host = '127.0.0.1'
        connect_port = local_addresses[0][1]
        if len(local_addresses) > 1 and 'oracle' == engine:
            local_addrs = ['(ADDRESS=(PROTOCOL=tcp)(HOST=127.0.0.1)(PORT=%d))' % (lh[1]) for lh in local_addresses]
            service_name = dbconf.get('ext_args', {}).get('service_name', '')
            if '' == service_name:
                service_name = dbconf.get('db', 'guest')
            # make oracle connect dsn
            connect_host = '(DESCRIPTION=(FAILOVER=on)(ADDRESS_LIST=%s)(CONNECT_DATA=(SERVICE_NAME=%s)))' % (' '.join(local_addrs), service_name)
            connect_port = 0
        return connect_host, connect_port, ssh_tunnel

    def get_default_db_port(self, engine: str) -> int:
        if 'oracle' == engine:
            return 1521
        elif 'postgresql' == engine:
            return 5432
        elif 'mssql' == engine:
            return 1433
        elif 'mysql' == engine:
            return 3306
        elif 'cockroach' == engine:
            return 26257
        elif 'mongodb' == engine:
            return 27017
        return 0

    def get_mongo_dbinstance(self, model) -> _DbInstance:
        """Get mongodb database instance by orm model
        :param model:modelutils.MongoBase implemented mongodb orm model
        :return: _DbInstance 
        """
        model_name = get_model_class_name(model)
        dbflag = get_dbinstance_by_model(model_name)
        if dbflag:
            if dbflag in self.db_instances:
                return self.db_instances[dbflag]
            else:
                LOG.error('query model:%s while could not determine the database instance', model_name)
                raise Exception('model %s configured db instance %s were not exists' % (model_name, dbflag))
        return self.default_mongo_db_instance

    def get_model_dbinstance(self, model) -> _DbInstance:
        """Get mongodb database instance by orm model
        :param model:modelutils.ModelBase sqlalchemy.ext.declarative.declarative_base implemented rdbms orm model
        
        :return: _DbInstance 
        """
        model_name = model if isinstance(model, str) else str(model._sa_class_manager.class_.__name__)
        dbflag = get_dbinstance_by_model(model_name)
        if dbflag:
            if dbflag in self.db_instances:
                return self.db_instances[dbflag]
            else:
                LOG.error('query model:%s while could not determine the database instance', model_name)
                raise Exception('model %s configured db instance %s were not exists' % (model_name, dbflag))
        return self.default_rdbms_db_instance

    def get_dbinstance(self, category: str) -> _DbInstance:
        """Get database instance by database category name
        :param category:str database category name configured by rdbms configurations
        
        :return: _DbInstance 
        """
        if category:
            if category in self.db_instances:
                return self.db_instances[category]
            else:
                LOG.error('could not get the database instance %s', category)
                raise Exception('db instance %s were not exists' % category)
        return self.default_rdbms_db_instance

    ################ part of rdbms operations ################
    # common queries
    async def query_list(self, model, filters, limit, offset, sort, direction, selections=None, joins=None, outerjoins=None, skipfields=None):
        """RDBMS query list by orm model
        :param model:modelutils.ModelBase sqlalchemy.ext.declarative.declarative_base implemented rdbms orm model
        :param filters:list|dict|tuple filter conditions
        :param limit:int returning rows limit
        :param offset:int rows offset in database
        :param sort:str sorting database table column
        :param direction:str sorting order, should be one of (asc|desc)
        :param selections:list select fields instead of all model fields
        :param joins:list multi table join condition
        :param outerjoins: list multi table outer join condition
        :param skipfields: replace DEFAULT_SKIP_FIELDS
        :return :list, int returns list of current queried rows and total records in database
        """
        qry, dbinstance, columns, _ = self._format_rdbms_query(model, filters, sort, direction, joins=joins, outerjoins=outerjoins)
        
        total = await self._execute_rdbms_query_count(dbinstance, qry)
        if not total:
            return [], total

        qry = qry.limit(limit).offset(offset)
        if selections:
            _selections = []
            for col in selections:
                if isinstance(col, str):
                    _selections.append(getattr(model, col))
                else:
                    _selections.append(col)
            qry = qry.with_entities(*_selections)
        rows = await self._execute_rdbms_result(dbinstance, qry, fetch_all=True)

        items = []
        if selections:
            for row in rows:
                item = {}
                cls_name_v_is_none = {}
                for k in selections:
                    col_name = k if isinstance(k, str) else k.key
                    if k.class_ == model:
                        item[col_name] = row._data[row._keymap[getattr(model, col_name).expression._label][1]]
                    else:
                        join_cls_name = k.class_._sa_class_manager.declarative_scan.classname
                        join_cls_k_value = row._data[row._keymap[getattr(k.class_, col_name).expression._label][1]]
                        cls_name_v_is_none[join_cls_name] = False
                        if join_cls_name not in item:
                            item[join_cls_name] = {
                                col_name: join_cls_k_value
                            }
                        else:
                            item[join_cls_name][col_name] = join_cls_k_value
                        if join_cls_k_value is not None:
                            cls_name_v_is_none[join_cls_name] = True
                for cls_name_k, cls_name_v in cls_name_v_is_none.items():
                    if not cls_name_v:
                        item[cls_name_k] = None
                items.append(item)
        else:
            for row in rows:
                item = {}
                for k in columns:
                    if isinstance(k, dict):
                        for kk, kv in k.items():
                            cls_name = kk._sa_class_manager.declarative_scan.classname
                            item[cls_name] = {}
                            cls_name_v_is_none = False
                            for kvk in kv:
                                kvv = row._data[row._keymap[getattr(kk, kvk).expression._label][1]]
                                if kvv is not None:
                                    cls_name_v_is_none = True
                                item[cls_name][kvk] = kvv
                            if not cls_name_v_is_none:
                                item[cls_name] = None
                        continue
                    if k in skipfields if skipfields is not None else DEFAULT_SKIP_FIELDS:
                        continue
                    item[k] = row._data[row._keymap[getattr(model, k).expression._label][1]]
                items.append(item)

        return items, total

    def _format_rdbms_query(self, model, filters, sort, direction, joins=None, outerjoins=None):
        columns,pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        engine_name = dbinstance.engine.dialect.name
        orderby = None
        if sort:
            if direction == 'desc':
                orderby = sqlalchemy.desc(getattr(model, sort))
            else:
                orderby = sqlalchemy.asc(getattr(model, sort))
        elif (engine_name == 'mssql' or engine_name == 'postgresql'):
            if pk:
                orderby = getattr(model, pk)
            else:
                orderby = getattr(model, columns[0])

        model_c = list(model._sa_class_manager.local_attrs.values())
        if joins:
            for join in joins:
                join_item_model_c = []
                for attr in join[0].__dict__.values():
                    if not isinstance(attr, sqlalchemy.orm.attributes.InstrumentedAttribute):
                        continue
                    join_item_model_c.append(attr)
                model_c.extend(join_item_model_c)
                join_columns, _ = model_columns(join[0])
                columns.append({
                    join[0]: join_columns
                })
        if outerjoins:
            for outerjoin in outerjoins:
                outerjoin_item_model_c = []
                for attr in outerjoin[0].__dict__.values():
                    if not isinstance(attr, sqlalchemy.orm.attributes.InstrumentedAttribute):
                        continue
                    outerjoin_item_model_c.append(attr)
                model_c.extend(outerjoin_item_model_c)
                outerjoin_columns, _ = model_columns(outerjoin[0])
                columns.append({
                    outerjoin[0]: outerjoin_columns
                })
        qry = query.Query(model_c)
        if joins:
            for join in joins:
                qry = qry.join(*join)
        if outerjoins:
            for outerjoin in outerjoins:
                qry = qry.outerjoin(*outerjoin)
        qry = qry.filter(*filters)
        if isinstance(orderby, list):
            qry = qry.order_by(*orderby)
        elif orderby is not None:
            qry = qry.order_by(orderby)
        return qry, dbinstance, columns, pk

    async def query_all(self, model, filters, sort=None, direction='asc', joins=None, skip_fields=None):
        qry, dbinstance, columns, _ = self._format_rdbms_query(model, filters, sort, direction, joins=joins)
        
        rows = await self._execute_rdbms_result(dbinstance, qry, fetch_all = True)
        items = []
        for row in rows:
            item = {}
            for k in columns:
                if not isinstance(k, str):
                    # 关联查询时，columns种存在dict类型的元素
                    continue
                this_skip_fields = DEFAULT_SKIP_FIELDS
                if skip_fields is not None:
                    this_skip_fields = skip_fields
                if k in this_skip_fields:
                    continue
                item[k] = row._data[row._keymap[getattr(model, k).expression._label][1]]
            items.append(item)

        return items

    async def _execute_rdbms_result(self, dbinstance: _DbInstance, qry: query.Query, fetch_all: bool = False, fetch_one: bool = False, execution_options: dict = sqlalchemy.util.EMPTY_DICT, sql_params = None):
        query_statement, query_params = self._format_query_statement(qry, sql_params)
        if dbinstance.disconnected:
            await dbinstance.manual_connect()
            if dbinstance.disconnected:
                LOG.error('execute query [%s] on db connection %s while the connection were not connected.', str(query_statement), dbinstance.name)
                raise Exception('Connection by %s were not connected' % dbinstance.name)
        ret = None
        start_ts = time.time()
        cur_trans = None
        try:
            # asyncio with threading like oracle, mssql:
            if dbinstance.async_by_thread:
                async with dbinstance.engine.begin() as conn:
                    async with conn.begin() as trans:
                        cur_trans = trans
                        if query_params is None:
                            cursor = await conn.execute(query_statement, execution_options=execution_options)
                        else:
                            cursor = await conn.execute(query_statement, parameters=query_params, execution_options=execution_options)
                        ret = await self._fetching_records(cursor, fetch_all, fetch_one)
            else:
                async with AsyncSession(dbinstance.engine) as session:
                    if query_params is None:
                        cursor = await session.execute(query_statement, execution_options=execution_options)
                    else:
                        cursor = await session.execute(query_statement, params=query_params, execution_options=execution_options)
                    ret = await self._fetching_records(cursor, fetch_all, fetch_one)
                    if query_statement.is_insert or query_statement.is_update or query_statement.is_delete:
                        await session.commit()

            execute_ts = time.time()
            if execute_ts - 1 > start_ts:
                LOG.warn('Slow query, execute query [%s] on db connection %s had taken too much time (%.2fs) on start time:[%s]', str(query_statement), dbinstance.name, execute_ts - start_ts, time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(start_ts)))
            fetched_ts = 0
            # if query_statement.is_insert or query_statement.is_update or query_statement.is_delete:
            # if cur_trans:
            #     await cur_trans.commit()
            #     cur_trans = None
            fetched_ts = time.time()
            if fetched_ts and (fetched_ts - 1 > execute_ts):
                LOG.warn('Slow query fetching, execute query [%s] on db connection %s fetching %s results had taken too much time (%.2fs) on fetching time:[%s]', str(query_statement), dbinstance.name, ('all' if fetch_all else 'one'), fetched_ts - execute_ts, time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(execute_ts)))
        except sqlalchemy.exc.OperationalError as e:
            LOG.error('query sql %s failed with error(%s):%s', str(query_statement), str(e.code), str(e))
            ExceptionReporter().report(key='SQL-'+str('query'), typ='SQL', 
                endpoint='%s|%s' % (str(dbinstance.name), str(query_statement)),
                method='QUERY',
                inputs=str(query_statement),
                outputs='',
                content=str(e),
                level='ERROR'
            )
            if cur_trans:
                LOG.info('begin rollback sql [%s]', str(query_statement))
                await cur_trans.rollback()
                cur_trans = None
                LOG.info('rollback sql [%s] finished', str(query_statement))
            if e.connection_invalidated:
                dbinstance.ondisconnected(str(e), True)
            raise e
        except sqlalchemy.exc.DatabaseError as e:
            LOG.error('query sql %s failed with error(%s):%s', str(query_statement), str(e.code), str(e))
            ExceptionReporter().report(key='SQL-'+str('query'), typ='SQL', 
                endpoint='%s|%s' % (str(dbinstance.name), str(query_statement)),
                method='QUERY',
                inputs=str(query_statement),
                outputs='',
                content=str(e),
                level='ERROR'
            )
            if cur_trans:
                LOG.info('begin rollback sql [%s]', str(query_statement))
                await cur_trans.rollback()
                cur_trans = None
                LOG.info('rollback sql [%s] finished', str(query_statement))
            if True or e.connection_invalidated:
                dbinstance.ondisconnected(str(e), True)
            raise e
        except Exception as e:
            LOG.error('query sql %s failed with error:%s', str(query_statement), str(e))
            ExceptionReporter().report(key='SQL-'+str('query'), typ='SQL', 
                endpoint='%s|%s' % (str(dbinstance.name), str(query_statement)),
                method='QUERY',
                inputs=str(query_statement),
                outputs='',
                content=str(e),
                level='ERROR'
            )
            if cur_trans:
                LOG.info('begin rollback sql [%s]', str(query_statement))
                await cur_trans.rollback()
                cur_trans = None
                LOG.info('rollback sql [%s] finished', str(query_statement))
            raise e
        finally:
            pass
        return ret

    def _format_query_statement(self, qry: query.Query, sql_params = None):
        query_statement = ''
        query_params = None
        if isinstance(qry, query.Query):
            querycontext = qry.with_labels()._compile_context()
            if hasattr(querycontext, 'statement'):
                # querycontext.statement.use_labels = True
                query_statement = querycontext.statement
            else:
                query_statement = querycontext.query
            query_params = qry._params
        elif isinstance(qry, str):
            query_statement = sqlalchemy.text(qry)
            query_params = sql_params
        else:
            query_statement = qry
            if hasattr(qry, '_params'):
                query_params = qry._params
        return query_statement, query_params

    async def _fetching_records(self, cursor, fetch_all: bool = False, fetch_one: bool = False):
        ret = None
        if cursor.returns_rows:
            if fetch_all:
                if asyncio.iscoroutinefunction(cursor.fetchall):
                    ret = await cursor.fetchall()
                else:
                    ret = cursor.fetchall()
            elif fetch_one:
                if asyncio.iscoroutinefunction(cursor.fetchone):
                    ret = await cursor.fetchone()
                else:
                    ret = cursor.fetchone()
            else:
                ret = cursor
        else:
            ret = cursor
        return ret

    async def _execute_rdbms_query_count(self, dbinstance: _DbInstance, qry: query.Query):
        col = sqlalchemy.sql.func.count(sqlalchemy.sql.literal_column("*"))
        qrycount = qry.from_self(col)
        # querycontext = qrycount._compile_context()
        # querycontext.statement.use_labels = True
        ret = await self._execute_rdbms_result(dbinstance, qrycount, fetch_one = True)
        if ret:
            return ret[0]
        return 0

    async def find_item(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        qry = query.Query(model).filter(*filters)
        row = await self._execute_rdbms_result(dbinstance, qry, fetch_one=True)
        if not row:
            return None
        item = model()
        columns, _ = model_columns(model)            
        for k in columns:
            setattr(item, k, row._data[row._keymap[getattr(model, k).expression._label][1]])
        return item

    async def get_count(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        qry = query.Query(model).filter(*filters)
        count = await self._execute_rdbms_query_count(dbinstance, qry)
        return count

    async def update_values(self, model, filters, values: dict):
        dbinstance = self.get_model_dbinstance(model)
        stmt = sqlalchemy.update(model).filter(*filters).values(**values)
        result = await self._execute_rdbms_result(dbinstance, stmt, execution_options={'synchronize_session': False})
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    async def update_item(self, item):
        model = item.__class__
        columns, pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        values, defaults = self.get_rdbms_instance_update_values(item, model, columns, pk)
        if not values:
            return False
        stmt = sqlalchemy.update(model).filter(getattr(model, pk)==getattr(item, pk)).values(**values)
        result = await self._execute_rdbms_result(dbinstance, stmt)
        if result and result.rowcount:
            for k, v in defaults.items():
                setattr(item, k, v)
            
        return item

    async def insert_item(self, item, auto_flush=False):
        model = item.__class__
        columns, pk = model_columns(model)
        dbinstance = self.get_model_dbinstance(model)
        values, defaults = self.get_rdbms_instance_insert_values(item, model, columns, pk)
        if not values:
            return False
        stmt = sqlalchemy.insert(model).values(**values)
        result = await self._execute_rdbms_result(dbinstance, stmt)
        if result and result.rowcount:
            for k, v in defaults.items():
                setattr(item, k, v)
            if result.inserted_primary_key:
                setattr(item, pk, result.inserted_primary_key[0])
            
        return item

    async def insert_items(self, items, auto_flush=False):
        if not items:
            return None
        insert_groups = {}
        for item in items:
            model = item.__class__
            if model not in insert_groups:
                insert_groups[model] = []
            insert_groups[model].append(item)
        for model, insert_group in insert_groups.items():
            columns, pk = model_columns(model)
            dbinstance = self.get_model_dbinstance(model)
            if 'oracle' == dbinstance.engine.name:
                try:
                    if len(insert_group) > 10:
                        LOG.warning('insert items of %s count:%d for %s that the engine does not support batch inserts, this would take a long time', str(model.__name__), len(insert_group), dbinstance.engine.name)
                    cur_group = []
                    insert_batches = []
                    i = 0
                    for item in insert_group:
                        cur_group.append(item)
                        i += 1
                        if i > 99:
                            insert_batches.append(cur_group)
                            cur_group = []
                            i = 0
                    if i > 0:
                        insert_batches.append(cur_group)
                    i = 1
                    for insert_batch in insert_batches:
                        async with AsyncSession(dbinstance.engine) as session:
                            LOG.info('inserting items of %s by count:%d on batch:%d', str(model.__name__), len(insert_batch), i)
                            t01 = time.time()
                            async with session.begin():
                                for item in insert_batch:
                                    session.add(item)
                            await session.commit()
                            t02 = time.time()
                            LOG.info('inserted items of %s by count:%d on batch:%d taken %.2f secs', str(model.__name__), len(insert_batch), i, t02 - t01)
                            i += 1
                except sqlalchemy.exc.OperationalError as e:
                    LOG.error('query insert items of %s failed with error(%s):%s', str(model.__name__), str(e.code), str(e))
                    ExceptionReporter().report(key='SQL-'+str('INSERT'), typ='SQL', 
                        endpoint='%s|%s' % (str(dbinstance.name), str(model.__name__)),
                        method='INSERT',
                        inputs=str(model.__name__),
                        outputs=str(e),
                        content=str(traceback.format_exc()),
                        level='ERROR'
                    )
                    if e.connection_invalidated:
                        dbinstance.ondisconnected(str(e), True)
                    raise e
                except sqlalchemy.exc.DatabaseError as e:
                    LOG.error('query insert items of %s failed with error(%s):%s', str(model.__name__), str(e.code), str(e))
                    ExceptionReporter().report(key='SQL-'+str('INSERT'), typ='SQL', 
                        endpoint='%s|%s' % (str(dbinstance.name), str(model.__name__)),
                        method='INSERT',
                        inputs=str(model.__name__),
                        outputs=str(e),
                        content=str(traceback.format_exc()),
                        level='ERROR'
                    )
                    if True or e.connection_invalidated:
                        dbinstance.ondisconnected(str(e), True)
                    raise e
                except Exception as e:
                    LOG.error('query insert items of %s failed with error:%s', str(model.__name__), str(e))
                    ExceptionReporter().report(key='SQL-'+str('INSERT'), typ='SQL', 
                        endpoint='%s|%s' % (str(dbinstance.name), str(model.__name__)),
                        method='INSERT',
                        inputs=str(model.__name__),
                        outputs=str(e),
                        content=str(traceback.format_exc()),
                        level='ERROR'
                    )
                    raise e
                continue
            total_idx = len(insert_group)
            if total_idx > 1000:
                start_idx = 0
                end_idx = 500
                LOG.warning('insert items of %s count:%d too large, split the operation in batch of %d', model.__name__, total_idx, end_idx)
                while end_idx <= total_idx:
                    insert_values = []
                    # insert_defaults = []
                    for i in range(start_idx, end_idx):
                        values, defaults = self.get_rdbms_instance_insert_values(insert_group[i], model, columns, pk)
                        insert_values.append(values)
                        # insert_defaults.append(defaults)

                    stmt = sqlalchemy.insert(model).values(insert_values)
                    result = await self._execute_rdbms_result(dbinstance, stmt)
                    # for i in range(start_idx, end_idx):
                    #     for k, v in insert_defaults[i].items():
                    #         setattr(insert_group[i], k, v)
                    #     if result.inserted_primary_key:
                    #         setattr(insert_group[i], pk, result.inserted_primary_key[i-start_idx])
                    
                    start_idx = end_idx
                    end_idx = end_idx + 500
                    if end_idx > total_idx:
                        end_idx = total_idx
                    if start_idx >= end_idx:
                        break
            else:
                insert_values = []
                # insert_defaults = []
                for i in range(0, total_idx):
                    values, defaults = self.get_rdbms_instance_insert_values(insert_group[i], model, columns, pk)
                    insert_values.append(values)
                    # insert_defaults.append(defaults)

                stmt = sqlalchemy.insert(model).values(insert_values)
                result = await self._execute_rdbms_result(dbinstance, stmt)
            
        return True

    async def del_item(self, item):
        model = item.__class__
        dbinstance = self.get_model_dbinstance(model)
        _, pk = model_columns(model)
        stmt = sqlalchemy.delete(model).where(getattr(model, pk)==getattr(item, pk))
        result = await self._execute_rdbms_result(dbinstance, stmt)
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    async def del_items(self, model, filters):
        dbinstance = self.get_model_dbinstance(model)
        stmt = sqlalchemy.delete(model).where(*filters)
        result = await self._execute_rdbms_result(dbinstance, stmt, execution_options={'synchronize_session': False})
        if result and result.rowcount:
            return result.rowcount
            
        return 0

    def get_rdbms_instance_insert_values(self, item, model, columns, pk):
        values = {}
        defaults = {}
        for col in columns:
            tbl_col_name = getattr(model, col).expression.name
            if col in item._sa_instance_state.unmodified:
                v = self.get_rdbms_instance_default_value(getattr(model, col).expression, True)
                if v is None and col == pk:
                    continue
                values[tbl_col_name] = v
                defaults[tbl_col_name] = v
            else:
                values[tbl_col_name] = getattr(item, col)
        return values, defaults

    def get_rdbms_instance_update_values(self, item, model, columns, pk):
        values = {}
        defaults = {}
        for col in columns:
            if col == pk:
                continue
            # tbl_col_name = getattr(model, col).expression.name
            values[col] = getattr(item, col)
            v = self.get_rdbms_instance_default_value(getattr(model, col).expression, False)
            if v is not None:
                values[col] = v
                defaults[col] = v
            
        return values, defaults

    def get_rdbms_instance_default_value(self, column, is_insert):
        column_default = column.default if is_insert else column.onupdate
        if column_default is None:
            return None
        else:
            return self._exec_rdbms_default(column, column_default, column.type)

    def _exec_rdbms_default(self, column, default, type_):
        if default.is_sequence:
            # raise Exception("sequence default column value is not supported")
            return None # self.fire_sequence(default, type_)
        elif default.is_callable:
            return default.arg(None)
        elif default.is_clause_element:
            # TODO: expensive branching here should be
            raise Exception("clause default column value is not supported")
            # # pulled into _exec_scalar()
            # conn = self.connection
            # if not default._arg_is_typed:
            #     default_arg = expression.type_coerce(default.arg, type_)
            # else:
            #     default_arg = default.arg
            # c = expression.select([default_arg]).compile(bind=conn)
            # return conn._execute_compiled(c, (), {}).scalar()
            return None
        else:
            return default.arg

    async def exec_query(self, db_category, sql, arguments: dict = None):
        """
        execute sql for orm dbs
        """
        if (not sql):
            return []
        dbinst: _DbInstance = self.get_dbinstance(db_category)
        # t1 = time.time()
        rows = await self._execute_rdbms_result(dbinst, sql, fetch_all=True, sql_params=arguments)
        # t2 = time.time()
        # print("------>>>>> %.2f executing [%s] from [%s] to [%s]" % (t2 - t1, str(sql), time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(t1)), time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(t2))))
        return rows

    async def exec_update(self, db_category, sql, arguments: dict = None):
        """
        execute sql for no back, update or insert
        """
        if (not sql):
            return False
        dbinstance: _DbInstance = self.get_dbinstance(db_category)
        sql_stmt = sqlalchemy.text(sql)
        sql_stmt.is_update = True
        await self._execute_rdbms_result(dbinstance, sql_stmt, execution_options={'synchronize_session': False}, sql_params=arguments)
        return True
    
    # call db procedure
    async def call_procedure(self, db_category, proc_name, params, out_params=None):
        # print("1")
        dbinstance: _DbInstance = self.get_dbinstance(db_category)
        if dbinstance.disconnected:
            await dbinstance.manual_connect()
            if dbinstance.disconnected:
                LOG.error('call procedure [%s] on db connection %s while the connection were not connected.', proc_name, db_category)
                raise Exception('Connection by %s were not connected' % db_category)
        if not isinstance(params, list):
            return False
        # print("2")
        conn = await dbinstance.engine.raw_connection()
        # print("3")
        cursor = conn.cursor()
        # print("4")
        result_values = {}
        if out_params and isinstance(out_params, dict) and 'oracle' == dbinstance.engine.sync_engine.name:
            for k, v in out_params.items():
                result_values[k] = cursor.var(getattr(cx_Oracle, v))
            params.extend(result_values.values())
        else:
            result_values = []
        # 调用存储过程
        # print("5")
        cursor.callproc(proc_name, params)
        # print("6")
        result = []
        if isinstance(result_values, dict):
            result = {k:int(result_values[k].getvalue()) if out_params.get(k) in ['NUMBER', 'INTEGER', 'INT'] and result_values[k].getvalue() is not None 
                else result_values[k].getvalue() for k in result_values}
        elif out_params:
            result = []
            for cursor_result in cursor.stored_results():
                result.append(cursor_result.fetchall())
        # print("7")
        cursor.close()
        # print("8")
        return result
    ################ end part of rdbms operations ################

    ################ part of mongodb operations ################
    async def query_list_mongo(self, model, filters, limit, offset, sort, direction, selections=None, joins=None, as_dict=True):
        q = self._format_mongo_query(model, filters)
        if selections:
            selFields = {}
            for k in selections:
                selFields[k] = 1
            if selFields:
                q = q.fields(*selFields)
        # if sort:
        #     if direction.lower() == 'desc':
        #         q = q.order_by('-'+sort)
        #     else:
        #         q = q.order_by(sort)
        if joins:
            # TODO
            pass

        collection = self._prepare_mongo_collection(model)
        total = await getattr(collection, self._motor_count_documents_name)(q._query)

        q = q.skip(offset)

        if sort:
            direc = pymongo.ASCENDING
            if direction.lower() == 'desc':
                direc = pymongo.DESCENDING
            cursor = collection.find(q._query).sort(sort, direc)
        else:
            cursor = collection.find(q._query)
        rows = await cursor.to_list(length=limit)

        items = []
        if as_dict:
            if selections:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: format_mongo_value(v) for k, v in row.items() if k in selections}
                    items.append(item)
            else:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: format_mongo_value(v) for k, v in row.items() if k not in DEFAULT_SKIP_FIELDS}
                    items.append(item)
        else:
            conditions = {}
            prepare_items = {}
            for row in rows:
                item = model()
                for k, v in row.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            if attr not in conditions:
                                conditions[attr] = {'filters': {pk+'__in': []}, 'ref_doc_type': ref_doc_type}
                            if attr not in prepare_items:
                                prepare_items[attr] = {v: []}
                            elif v not in prepare_items[attr]:
                                prepare_items[attr][v] = []
                            conditions[attr]['filters'][pk+'__in'].append(v)
                            prepare_items[attr][v].append(item)
                    
                    setattr(item, attr, v)
                items.append(item)

            if conditions:
                for attr, ref_options in conditions.items():
                    ref_rows = await self.query_all_mongo(ref_options['ref_doc_type'], ref_options['filters'], limit=limit, as_dict=as_dict)
                    for ref_row in ref_rows:
                        if ref_row.id in prepare_items[attr]:
                            for dst_item in prepare_items[attr][ref_row.id]:
                                setattr(dst_item, attr, ref_row)
        
        return items, total

    async def query_all_mongo(self, model, filters, limit=100, sort=None, **kwargs):
        direction = kwargs.pop('direction', None)
        selections = kwargs.pop('selections', None)
        joins = kwargs.pop('joins', None)
        as_dict = kwargs.pop('as_dict', True)
        q = self._format_mongo_query(model, filters)
        if selections:
            selFields = {}
            for k in selections:
                selFields[k] = 1
            if selFields:
                q = q.fields(*selFields)
        # if sort:
        #     if direction.lower() == 'desc':
        #         q = q.order_by('-'+sort)
        #     else:
        #         q = q.order_by(sort)
        if joins:
            # TODO
            pass

        collection = self._prepare_mongo_collection(model)

        if sort:
            direc = pymongo.ASCENDING
            if direction.lower() == 'desc':
                direc = pymongo.DESCENDING
            cursor = collection.find(q._query).sort(sort, direc)
        else:
            cursor = collection.find(q._query)
        rows = await cursor.to_list(length=limit)

        items = []
        if as_dict:
            if selections:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in row.items() if k in selections}
                    items.append(item)
            else:
                for row in rows:
                    item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in row.items() if k not in DEFAULT_SKIP_FIELDS}
                    items.append(item)
        else:
            conditions = {}
            prepare_items = {}
            for row in rows:
                item = model()
                for k, v in row.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            if attr not in conditions:
                                conditions[attr] = {'filters': {pk+'__in': []}, 'ref_doc_type': ref_doc_type}
                            if attr not in prepare_items:
                                prepare_items[attr] = {v: []}
                            elif v not in prepare_items[attr]:
                                prepare_items[attr][v] = []
                            conditions[attr]['filters'][pk+'__in'].append(v)
                            prepare_items[attr][v].append(item)
                    
                    setattr(item, attr, v)
                items.append(item)

            if conditions:
                for attr, ref_options in conditions.items():
                    ref_rows = await self.query_all_mongo(ref_options['ref_doc_type'], ref_options['filters'], limit=limit, as_dict=as_dict)
                    for ref_row in ref_rows:
                        if ref_row.id in prepare_items[attr]:
                            for dst_item in prepare_items[attr][ref_row.id]:
                                setattr(dst_item, attr, ref_row)

        return items

    async def find_one_mongo(self, model, *args, **kwargs):
        """
        """
        as_dict = kwargs.pop('as_dict', False)
        q = self._format_mongo_query(model, (args, kwargs))
        collection = self._prepare_mongo_collection(model)
        document = await collection.find_one(q._query)
        if not document:
            return None

        if as_dict:
            item = {model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k: v for k, v in document.items()}
            # for k1, k2 in model._reverse_db_field_map.items():
            #     item[k2] = format_mongo_value(document.get(k1))
            return item
        else:
            item = model()
            for k, v in document.items():
                attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                    field = model._fields.get(attr)
                    ref_doc_type = field.document_type
                    pk = ref_doc_type._reverse_db_field_map.get('_id')
                    if pk:
                        ref_filters = {pk: v, 'as_dict': as_dict}
                        v = await self.find_one_mongo(ref_doc_type, **ref_filters)

                setattr(item, attr, v)

        return item

    async def mongo_aggregate(self, model, pipeline=[], *args, **kwargs):
        """
        query mongo by aggregate.
        """
        as_dict = kwargs.pop('as_dict', True)
        collection = self._prepare_mongo_collection(model)
        cursor = collection.aggregate(pipeline)
        data_list = []
        if as_dict:
            while (await cursor.fetch_next):
                item = cursor.next_object()
                data_list.append(item)
        else:
            while (await cursor.fetch_next):
                item = cursor.next_object()
                obj = model()
                for k, v in item.items():
                    attr = model._reverse_db_field_map[k] if k in model._reverse_db_field_map else k
                    if isinstance(model._fields.get(attr), mongoengine.fields.ReferenceField):
                        field = model._fields.get(attr)
                        ref_doc_type = field.document_type
                        pk = ref_doc_type._reverse_db_field_map.get('_id')
                        if pk:
                            ref_filters = {pk: v, 'as_dict': as_dict}
                            v = await self.find_one_mongo(ref_doc_type, **ref_filters)

                    setattr(obj, attr, v)
                data_list.append(obj)

        return data_list

    async def save_mongo(self, item, force_insert=False, validate=True, clean=True,
             write_concern=None, cascade=None, cascade_kwargs=None,
             _refs=None, save_condition=None, signal_kwargs=None, **kwargs):
        """Save the :class:`~mongoengine.Document` to the database. If the
        document already exists, it will be updated, otherwise it will be
        created.

        :param force_insert: only try to create a new document, don't allow
            updates of existing documents.
        :param validate: validates the document; set to ``False`` to skip.
        :param clean: call the document clean method, requires `validate` to be
            True.
        :param write_concern: Extra keyword arguments are passed down to
            :meth:`~pymongo.collection.Collection.save` OR
            :meth:`~pymongo.collection.Collection.insert`
            which will be used as options for the resultant
            ``getLastError`` command.  For example,
            ``save(..., write_concern={w: 2, fsync: True}, ...)`` will
            wait until at least two servers have recorded the write and
            will force an fsync on the primary server.
        :param cascade: Sets the flag for cascading saves.  You can set a
            default by setting "cascade" in the document __meta__
        :param cascade_kwargs: (optional) kwargs dictionary to be passed throw
            to cascading saves.  Implies ``cascade=True``.
        :param _refs: A list of processed references used in cascading saves
        :param save_condition: only perform save if matching record in db
            satisfies condition(s) (e.g. version number).
            Raises :class:`OperationError` if the conditions are not satisfied
        :param signal_kwargs: (optional) kwargs dictionary to be passed to
            the signal calls.

        .. versionchanged:: 0.5
            In existing documents it only saves changed fields using
            set / unset.  Saves are cascaded and any
            :class:`~bson.dbref.DBRef` objects that have changes are
            saved as well.
        .. versionchanged:: 0.6
            Added cascading saves
        .. versionchanged:: 0.8
            Cascade saves are optional and default to False.  If you want
            fine grain control then you can turn off using document
            meta['cascade'] = True.  Also you can pass different kwargs to
            the cascade save using cascade_kwargs which overwrites the
            existing kwargs with custom values.
        .. versionchanged:: 0.8.5
            Optional save_condition that only overwrites existing documents
            if the condition is satisfied in the current db record.
        .. versionchanged:: 0.10
            :class:`OperationError` exception raised if save_condition fails.
        .. versionchanged:: 0.10.1
            :class: save_condition failure now raises a `SaveConditionError`
        .. versionchanged:: 0.10.7
            Add signal_kwargs argument
        """
        if item._meta.get('abstract'):
            raise mongoengine.InvalidDocumentError('Cannot save an abstract document.')

        signal_kwargs = signal_kwargs or {}
        mongoengine.signals.pre_save.send(item.__class__, document=item, **signal_kwargs)

        if validate:
            item.validate(clean=clean)

        if write_concern is None:
            write_concern = {'w': 1}

        doc = item.to_mongo()

        created = ('_id' not in doc or item._created or force_insert)

        mongoengine.signals.pre_save_post_validation.send(item.__class__, document=item,
                                              created=created, **signal_kwargs)
        # it might be refreshed by the pre_save_post_validation hook, e.g., for etag generation
        doc = item.to_mongo()

        if item._meta.get('auto_create_index', True):
            await self.mongo_model_ensure_indexes(item)

        try:
            # Save a new document or update an existing one
            if created:
                object_id = await self._mongo_save_create(item, doc, force_insert, write_concern)
            else:
                object_id, created = await self._mongo_save_update(item, doc, save_condition,
                                                                   write_concern
                                                                  )

            if cascade is None:
                cascade = (item._meta.get('cascade', False) or
                           cascade_kwargs is not None)

            if cascade:
                kwargs = {
                    'force_insert': force_insert,
                    'validate': validate,
                    'write_concern': write_concern,
                    'cascade': cascade
                }
                if cascade_kwargs:  # Allow granular control over cascades
                    kwargs.update(cascade_kwargs)
                kwargs['_refs'] = _refs
                await self.mongo_cascade_save(item, **kwargs)

        except pymongo.errors.DuplicateKeyError as err:
            message = 'Tried to save duplicate unique keys (%s)'
            raise mongoengine.NotUniqueError(message % six.text_type(err))
        except pymongo.errors.OperationFailure as err:
            message = 'Could not save document (%s)'
            if re.match('^E1100[01] duplicate key', six.text_type(err)):
                # E11000 - duplicate key error index
                # E11001 - duplicate key on update
                message = 'Tried to save duplicate unique keys (%s)'
                raise mongoengine.NotUniqueError(message % six.text_type(err))
            raise mongoengine.OperationError(message % six.text_type(err))

        # Make sure we store the PK on this document now that it's saved
        id_field = item._meta['id_field']
        if created or id_field not in item._meta.get('shard_key', []):
            item[id_field] = item._fields[id_field].to_python(object_id)

        mongoengine.signals.post_save.send(item.__class__, document=item,
                               created=created, **signal_kwargs)

        item._clear_changed_fields()
        item._created = False

        return item

    async def insert_mongo(self, model, values):
        """Supports Document, dict, or list(builk insert)

        :param mongoengine.Document|str model: The mongo schema or collection name.
        :param mongoengine.Document|dict|list values: the mongo document or dict or list value to be inserted.
        :return bool:
        """
        collection = self._prepare_mongo_collection(model)
        result = False
        if isinstance(values, mongoengine.Document):
            result = await self.save_mongo(values)
        elif isinstance(values, dict):
            result = await collection.insert_one(values)
        else:
            inserts = []
            for v in values:
                if isinstance(v, dict):
                    inserts.append(v)
                elif isinstance(v, mongoengine.Document):
                    item = v.to_mongo()
                    inserts.append(item)
                else:
                    raise Exception("insert mongodb values should either be dict or Document type!")
            if inserts:
                result = await collection.insert_many(inserts)
        return result

    async def update_mongo(self, model, item, condition):
        """
        update mongo model

        :param mongoengine.Document|str model: The mongo schema or collection name.
        :param dict item: update key and value.
        :param dict condition: update by this conditions.
        :return int: updated count.
        """
        if not (condition and isinstance(item, dict) and ("id" or "_id") not in item):
            raise mongoengine.OperationError("Invalid input.")
        collection = self._prepare_mongo_collection(model)

        try:
            updated_ret = await collection.update_many(condition, {"$set": item})
            updated_count = updated_ret.modified_count
        except Exception:
            updated_count = 0

        return updated_count

    async def del_item_mongo(self, item, condition=None):
        """
        :param mongoengine.Document|str item: The mongo document or schema or collection name
        :param dict|None condition: the delete condition if condition specified, the field 
            should be provided when item is schame or collection name.
        :return bool:
        """
        result = False
        collection = self._prepare_mongo_collection(item)
        if not condition:
            result = await collection.delete_one({'_id': item.pk})
        elif isinstance(condition, dict):
            result = await collection.delete_many(condition)
        return result

    def _prepare_mongo_collection(self, model):
        dbinst = self.get_mongo_dbinstance(model)
        if isinstance(model, str):
            collection_name = model
        else:
            collection_name = model._meta.get('collection')
            if not collection_name:
                collection_name = model._class_name
        collection = dbinst.conn[collection_name]
        return collection

    def _format_mongo_query(self, model, filters):
        q = mongoengine.queryset.QuerySet(model, None)
        qfilters = []
        kwfilters = {}
        if isinstance(filters, tuple):
            for f in filters:
                if isinstance(f, dict):
                    for k,v in f.items():
                        kwfilters[k] = v
                elif isinstance(f, list):
                    for v in f:
                        qfilters.append(v)
        elif isinstance(filters, dict):
            kwfilters = filters
        elif isinstance(filters, list):
            qfilters = filters

        if qfilters or kwfilters:
            q = q.filter(*qfilters, **kwfilters)
        elif filters:
            q = q.filter(filters)
        return q

    async def mongo_model_ensure_indexes(self, model):
        """Checks the document meta data and ensures all the indexes exist.

        Global defaults can be set in the meta - see :doc:`guide/defining-documents`

        .. note:: You can disable automatic index creation by setting
                  `auto_create_index` to False in the documents meta data
        """
        return
        background = model._meta.get('index_background', False)
        drop_dups = model._meta.get('index_drop_dups', False)
        index_opts = model._meta.get('index_opts') or {}
        index_cls = model._meta.get('index_cls', True)

        collection = self._prepare_mongo_collection(model)
        # 746: when connection is via mongos, the read preference is not necessarily an indication that
        # this code runs on a secondary
        if not collection.is_mongos and collection.read_preference > 1:
            return

        # determine if an index which we are creating includes
        # _cls as its first field; if so, we can avoid creating
        # an extra index on _cls, as mongodb will use the existing
        # index to service queries against _cls
        cls_indexed = False

        # Ensure document-defined indexes are created
        if model._meta['index_specs']:
            index_spec = model._meta['index_specs']
            for spec in index_spec:
                spec = spec.copy()
                fields = spec.pop('fields')
                cls_indexed = cls_indexed or mongoengine.document.includes_cls(fields)
                opts = index_opts.copy()
                opts.update(spec)

                # we shouldn't pass 'cls' to the collection.ensureIndex options
                # because of https://jira.mongodb.org/browse/SERVER-769
                if 'cls' in opts:
                    del opts['cls']

                if hasattr(collection,"ensure_index") and  hasattr(mongoengine.pymongo_support,"IS_PYMONGO_GTE_37") and mongoengine.pymongo_support.IS_PYMONGO_GTE_37:
                    collection.create_index(fields, background=background, **opts)
                else:
                    if hasattr(collection,"ensure_index"):
                        collection.ensure_index(fields, background=background,
                                                drop_dups=drop_dups, **opts)

        # If _cls is being used (for polymorphism), it needs an index,
        # only if another index doesn't begin with _cls
        if index_cls and not cls_indexed and model._meta.get('allow_inheritance'):

            # we shouldn't pass 'cls' to the collection.ensureIndex options
            # because of https://jira.mongodb.org/browse/SERVER-769
            if 'cls' in index_opts:
                del index_opts['cls']

            if mongoengine.pymongo_support.IS_PYMONGO_GTE_37:
                collection.create_index('_cls', background=background,
                                        **index_opts)
            else:
                collection.ensure_index('_cls', background=background,
                                        **index_opts)

    async def _mongo_save_create(self, item, doc, force_insert, write_concern):
        """Save a new document.

        Helper method, should only be used inside save().
        """
        collection = self._prepare_mongo_collection(item)
        if force_insert:
            insert_result = await collection.insert_one(doc)
            return insert_result.inserted_id
        # insert_one will provoke UniqueError alongside save does not
        # therefore, it need to catch and call replace_one.
        if '_id' in doc:
            raw_object = await collection.replace_one(
                {'_id': doc['_id']}, doc)
            if raw_object:
                return doc['_id']

        insert_result = await collection.insert_one(doc)
        object_id = insert_result.inserted_id

        # In PyMongo 3.0, the save() call calls internally the _update() call
        # but they forget to return the _id value passed back, therefore getting it back here
        # Correct behaviour in 2.X and in 3.0.1+ versions
        if not object_id and pymongo.version_tuple == (3, 0):
            pk_as_mongo_obj = item._fields.get(item._meta['id_field']).to_mongo(item.pk)
            rc = await self.find_one_mongo(item.__class__, pk=pk_as_mongo_obj)
            if rc:
                object_id = rc.pk
            # object_id = (
            #     item._qs.filter(pk=pk_as_mongo_obj).first() and
            #     item._qs.filter(pk=pk_as_mongo_obj).first().pk
            # )  # TODO doesn't this make 2 queries?

        return object_id

    async def _mongo_save_update(self, item, doc, save_condition, write_concern):
        """Update an existing document.

        Helper method, should only be used inside save().
        """
        collection = self._prepare_mongo_collection(item)
        object_id = doc['_id']
        created = False

        select_dict = {}
        if save_condition is not None:
            select_dict = mongoengine.queryset.query(item.__class__, **save_condition)
            # select_dict = mongoengine.transform.query(item.__class__, **save_condition)

        select_dict['_id'] = object_id

        # Need to add shard key to query, or you get an error
        shard_key = item._meta.get('shard_key', tuple())
        for k in shard_key:
            path = item._lookup_field(k.split('.'))
            actual_key = [p.db_field for p in path]
            val = doc
            for ak in actual_key:
                val = val[ak]
            select_dict['.'.join(actual_key)] = val

        update_doc = item._get_update_doc()
        if update_doc:
            wc = write_concern.pop('w')
            upsert = save_condition is None
            last_error = await collection.update_one(select_dict, update_doc,
                                                     upsert=upsert, **write_concern)
            if not upsert and last_error._UpdateResult__raw_result['n'] == 0:
                raise mongoengine.SaveConditionError('Race condition preventing'
                                         ' document update detected')
            if last_error is not None:
                updated_existing = last_error.raw_result.get('updatedExisting')
                if updated_existing is False:
                    created = True
                    # !!! This is bad, means we accidentally created a new,
                    # potentially corrupted document. See
                    # https://github.com/MongoEngine/mongoengine/issues/564

        return object_id, created

    async def mongo_cascade_save(self, item, doc, **kwargs):
        """Recursively save any references and generic references on the
        document.
        """
        _refs = kwargs.get('_refs') or []

        ReferenceField = mongoengine.document._import_class('ReferenceField')
        GenericReferenceField = mongoengine.document._import_class('GenericReferenceField')

        for name, cls in list(item._fields.items()):
            if not isinstance(cls, (ReferenceField,
                                    GenericReferenceField)):
                continue

            ref = item._data.get(name)
            if not ref or isinstance(ref, mongoengine.DBRef):
                continue

            if not getattr(ref, '_changed_fields', True):
                continue

            ref_id = "%s,%s" % (ref.__class__.__name__, str(ref._data))
            if ref and ref_id not in _refs:
                _refs.append(ref_id)
                kwargs["_refs"] = _refs
                await self.save_mongo(ref, **kwargs)
                ref._changed_fields = []

    ################ end part of mongodb operations ################


#!/usr/bin/env python
# -*- coding: utf-8 -*-


from bson import ObjectId
import time, datetime
import mongoengine
from sqlalchemy.ext.declarative import declarative_base

ModelBase = declarative_base()

def model_columns(model):
    if isinstance(model, mongoengine.Document) or hasattr(model, '_reverse_db_field_map'):
        columns = [k for k in model._fields]
        pk = model._reverse_db_field_map.get('_id')
        return columns, pk
    columns = []
    pk = None
    cls_ = model._sa_class_manager.class_
    ref = model._sa_class_manager.deferred_scalar_loader.args[0]
    colmaps = {}
    for k in model._sa_class_manager._all_key_set:
        c = getattr(cls_, k)
        colmaps[c.expression.key] = c.key
    tbl = None
    if ref.primary_key:
        pk = colmaps[ref.primary_key[0].name]
    for t in ref.tables:
        if t.name == ref.local_table.name:
            tbl = t
            break
    if tbl is not None:
        for k in tbl.columns._all_columns:
            columns.append(colmaps[str(k.key)])
    else:
        for k in model._sa_class_manager._all_key_set:
            columns.append(str(k))
    return columns, pk

class MongoBase(mongoengine.DynamicDocument, metaclass=mongoengine.base.TopLevelDocumentMetaclass):
    """
    mongo模型基类
    """
    my_metaclass = mongoengine.base.TopLevelDocumentMetaclass

    def __init__(self, *args, **values):
        super(MongoBase, self).__init__(*args, **values)
        if not hasattr(self, 'obsoleted'):
            setattr(self, 'obsoleted', False)
        if not hasattr(self, 'created_at'):
            setattr(self, 'created_at', int(time.time()*1000))
        if not hasattr(self, 'updated_at'):
            setattr(self, 'updated_at', int(time.time()*1000))
        if not hasattr(self, 'sort_value'):
            setattr(self, 'sort_value', 1)
        # if not hasattr(self, 'created_by'):
        #     setattr(self, 'created_by',  None)
        # if not hasattr(self, 'updated_by'):
        #     setattr(self, 'updated_by',  None)

    def set_created(self, userId):
        self.created_by = userId

    def set_updated(self, userId):
        self.updated_by = userId

    def as_dict(self):
        result = {}
        for k in self:
            v = self[k]
            if isinstance(v, object):
                v = str(v)
            elif isinstance(v, datetime.datetime):
                v = str(v)
            result[k] = v
        return result

def format_mongo_value(v):
    if isinstance(v, ObjectId):
        return str(v)
    elif isinstance(v, mongoengine.Document):
        return {f: format_mongo_value(getattr(v, f)) for f in v}
    elif isinstance(v, datetime.datetime):
        return str(v)
    return v

MODEL_DB_MAPPING = {}
DEFAULT_SKIP_FIELDS = {'obsoleted':True, 'created_at':True, 'updated_at':True, 'created_by':True, 'updated_by':True}

def get_dbinstance_by_model(model_name):
    if model_name in MODEL_DB_MAPPING:
        return MODEL_DB_MAPPING[model_name]
    return ''

def set_model_dbinstance(model_instance_mapping: dict):
    if isinstance(model_instance_mapping, dict):
        for k, v in model_instance_mapping.items():
            MODEL_DB_MAPPING[k] = str(v)

def get_model_class_name(model) -> str:
    if hasattr(model, '_sa_class_manager'):
        return str(model._sa_class_manager.class_.__name__)
    elif hasattr(model, '__name__'):
        return model.__name__
    elif isinstance(model, str):
        return model
    return model.__class__.__name__



# -*- coding: utf-8 -*-
# @Author	: starview.brotherbaby
# @Date		: 2020-09-08 22:59:41 
# @Last Modified by:   starview.brotherbaby
# @Last Modified time: 2020-09-08 22:59:41
# Thanks for your comments!

from IPy import IP

def reload_white_ip_list(conf):
    whiltelist = {}
    if not conf:
        return whiltelist
    ips = conf.get('whitelist')
    if not ips:
        return whiltelist
    for ip in ips:
        ipx = IP(ip)
        for x in ipx:
            whiltelist[str(x)] = True
    return whiltelist

def is_rejected_by_white_list(whitelist, ip):
    if whitelist and ip in whitelist:
        return False
    return True



#!/usr/bin/env python
# encoding: utf-8
from lycium.kafka.helper import Helper
from lycium.kafka.asyncWrapper import AsyncWrapper
# from lycium.kafka.logger import logger
from confluent_kafka import Producer,Consumer,KafkaError,KafkaException,TopicPartition
import json
import asyncio
import sys
from lycium.kafka.protocol.kafkapacket_pb2 import KafkaPacket
import logging

logger = logging.getLogger('kafka')


class ConsumerHelper(Helper):
    """
    消费者辅助
    """
    def __init__(self):
        super(ConsumerHelper, self).__init__()
        self._config["group.id"]= 0
        self._config['max.poll.interval.ms']= 60 * 1000
        self._config["session.timeout.ms"] = 6000
        self._config["heartbeat.interval.ms"] = 2000
        self._consumer = None
        self._partition = 0
        # 值的解码器
        self._value_deserializer = None

        # key 的解码器
        self._key_deserializer = None

        # protobuf 解码时需要
        self._objectParser = None
        # 发送消息的状态，True 是成功发送，False 是发送失败
        self._send_error = None
        self._send_msg = None
        # 最小同步消息数量
        self._minCommitCount = 100
        # 记录消息的偏移量,当kafka集群中某个节点(是使用topic的Leader节点)出现问题挂掉,此时还能接收到消息。
        # 有可能出现在节点恢复后，重发节点挂掉后发送的消息,所以通过记录偏移量避免重复处理消息
        self._offset = -1

        self._offsetDict={}
        self._loop  = asyncio.get_event_loop()
        self._is_debug = False

        # 是否停止消费标记z
        self._stop_consume_dict = {}

        self._poll_tasks = []
        # self._loop.call_later(0.0001,self._loop.create_task,self._poll_task())

    def open_debug(self):
        self._is_debug = True

    async def _poll_task(self):
        for task in self._poll_tasks:
            fn,params = task
            await fn(params)
        self._loop.call_later(0.0001,self._loop.create_task,self._poll_task())

    async def poll_task(self):
        for task in self._poll_tasks:
            fn,params = task
            await fn(params)

    def config_group_id(self,groupID):
        """
        配置groupID
        """
        self._config["group.id"] = groupID
        return self

    def config_max_poll_interval_ms(self,interval=60 * 1000):
        """ 
        配置 poll(拉取数据)的间隔，默认是设置60 秒
        """
        self._config['max.poll.interval.ms']=  interval
        

    def config_value_deserializer(self,method:str ="protobuf",object_parser=KafkaPacket):
        """
        配置对值的解码方式，可选有string、json和protobuf 三种方法
        当使用protobuf 时候需要指定objectParser 为需要转换成的对象,注意objectParser 是一个类
        """
        if method == "string":
            self._value_deserializer=bytes.decode
        if method == "json":
            self._value_deserializer = self._deserializer_json
        if method == "protobuf":
            self._value_deserializer = self._deserializer_protobuf
            self._object_parser = object_parser
        # self._config['value_deserializer']= self._value_deserializer
        return self


    def receive(self,topics:list,work):
        """
        订阅一系列主题，work 是处理接收到消息的方法
        此方法是阻塞的方式运行
        """
        if self._consumer is None:
            self._consumer = Consumer(self._config)
        try:
            
            partitions = [TopicPartition(t,self._partition) for t in topics]
            self._consumer.assign(partitions)
            self._consumer.subscribe(topics)
            msgCount = 0
            while 1:
                msg = self._consumer.poll(timeout=1.10)
                
                if msg is None:
                    continue
                if msg.error():
                    print("error here")
                    print(msg.error())
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        # 读取到分区的末尾了
                        sys.stderr.write('%% %s [%d] reached end at offset %d\n' %
                                            (msg.topic(), msg.partition(), msg.offset()))
                    elif msg.error():
                        raise KafkaException(msg.error())
                else:
                    if msg.offset() > self._offset:
                        self._offset = msg.offset()
                        value =self._value_deserializer(msg.value())
                        work(value)
                    # msgCount += 1
                    # if msgCount % self._minCommitCount == 0:
                    #     pass
                        self._consumer.commit()
                    else:
                        print("skipping msg offset {0}".format(msg.offset()))
        finally:
            self._consumer.close()

    def receive_async(self,topics:list,work):
        """ 
        订阅一系列主题，work 是处理接收到消息的方法
        此方法是异步执行
        """
        async def receive_message(t):
            c_async,work,key = t
            if not self._stop_consume_dict[key]:
                msg = await c_async.poll(timeout=0.01)
                if msg is None:
                    return
                if msg.error():
                    print(msg.error())
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        # 读取到分区的末尾了
                        sys.stderr.write('%% %s [%d] reached end at offset %d\n' %
                                            (msg.topic(), msg.partition(), msg.offset()))
                else:
                    if self._is_debug:
                        logger.debug("topic:{0},partition:{1},offset:{2},msg value:{3}".format(
                            msg.topic(), 
                            msg.partition(), 
                            msg.offset(),
                            msg.value()
                        ))
                    offset_key = "{0}-{1}".format(key,msg.partition())
                    if offset_key not in self._offsetDict:
                        self._offsetDict[offset_key] = -1
                    if msg.offset() > self._offsetDict[offset_key]:
                        self._offsetDict[offset_key] = msg.offset()
                        value = self._value_deserializer(msg.value())
                        if asyncio.iscoroutine(work) or asyncio.iscoroutinefunction(work):
                            await work(value)
                        else:
                            work(value)
                        # c_async.commit()
                
            else:
                # 停止消费(接收信息)，关闭连接
                try:
                    if self._consumer is not None:
                        self._consumer.close()
                except Exception as e:
                    logger.exception(e)

        key="_".join(topics)
        self._stop_consume_dict[key] = False
        c = Consumer(self._config)
        partitions = [TopicPartition(t,self._partition) for t in topics]
        c.assign(partitions)
        c.subscribe(topics)
        c_async = AsyncWrapper(c,methods=["poll"])
        # self._offsetDict[c_async] = -1
        self._poll_tasks.append((receive_message,(c_async,work,key)))
        self._loop.call_later(0.00001,self._loop.create_task,self._poll_task())


    def stop_consumer(self,topics:list):
        """ 
        停止接收信息
        """
        
        key="_".join(topics)
        logger.info("{0} 停止接收信息".format(key))
        self._stop_consume_dict[key] = True
            
    def _deserializer_json(self,v):
        """
        把bytes 转dict
        """
        d = json.loads(v.decode("utf-8"))
        self._json_bytes_recover(d)
        return d

    def _json_bytes_recover(self,d):
        """
        对字典里面的字节数组进行恢复
        """
        to_remove_keys = []
        for k in d:
            if k.startswith("_c_"):
                update_key=k[3:]
                d[update_key]=d[update_key].encode("utf8")
                to_remove_keys.append(k)
            if isinstance(d[k],dict):
                self._json_bytes_recover(d[k])
        for k in to_remove_keys:
            del d[k]

    def _deserializer_protobuf(self,v):
        """
        把bytes 转成protobuf
        """
        op = self._object_parser()
        if '_register_private' in str(v):
            op = None
        else:
            op.ParseFromString(v)
        return op

    def config_message_size(self,size):
        self._config['fetch.message.max.bytes'] = size



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import time
import tornado.gen
from bson import ObjectId
from IPy import IP
from .supports import singleton
from .utilities import tofloat
from .utilities.signatureutils import verify_inputs_data_sign_md5, verify_inputs_data_sign_hmac
from .asynchttphandler import args_as_dict, request_body_as_json
from .dbproxy import DbProxy

LOG = logging.getLogger('components.accesscontrol')

@singleton
class AccessControlManager(object):
    """
    """
    def __init__(self):
        self.white_list = {}
        self.black_list = {}
        self.signatures = {}
        self.applications = {}

    @tornado.gen.coroutine
    def reload_from_mongo(self, config_model):
        yield self.reload_access_limit_from_mongo(config_model)
        yield self.reload_signature_from_mongo(config_model)
        
    @tornado.gen.coroutine
    def reload_access_limit_from_mongo(self, config_model):
        db_proxy = DbProxy()
        limit = 100
        offset = 0
        last_id = None
        n = limit
        filters = {'type': 'access-limit', 'obsoleted': False}
        while n >= limit:
            n = 0
            if last_id:
                filters['id__gt'] = last_id
            LOG.info('loading app_config by %s from db offset:%d', filters['type'], offset)
            rows, total = yield db_proxy.query_list_mongo(config_model, filters, limit, 0, 'id', 'asc')
            for row in rows:
                n = n + 1
                last_id = ObjectId(row['id'])
                app_id = row.get('app_id')
                if not app_id:
                    continue
                opts = row.get('options')
                if not opts:
                    continue

                if app_id not in self.white_list:
                    self.white_list[app_id] = {}
                if app_id not in self.black_list:
                    self.black_list[app_id] = {}
                white_list = self.white_list[app_id]
                black_list = self.black_list[app_id]
                wl = opts.get('whitelist')
                bl = opts.get('blacklist')
                if wl:
                    for ip in wl:
                        try:
                            ipx = IP(ip)
                            for x in ipx:
                                white_list[str(x)] = True
                        except Exception as e:
                            LOG.warning('loading app_id:%s white list for ip:%s failed with error:%s', app_id, str(ip), str(e))
                    for ip in bl:
                        try:
                            ipx = IP(ip)
                            for x in ipx:
                                black_list[str(x)] = True
                        except Exception as e:
                            LOG.warning('loading app_id:%s black list for ip:%s failed with error:%s', app_id, str(ip), str(e))
            offset += n
        LOG.info('loading app_config by %s finished.', filters['type'])
        
    @tornado.gen.coroutine
    def reload_signature_from_mongo(self, config_model):
        db_proxy = DbProxy()
        limit = 100
        offset = 0
        last_id = None
        n = limit
        filters = {'type': 'signature', 'obsoleted': False}
        while n >= limit:
            n = 0
            if last_id:
                filters['id__gt'] = last_id
            LOG.info('loading app_config by %s from db offset:%d', filters['type'], offset)
            rows, total = yield db_proxy.query_list_mongo(config_model, filters, limit, 0, 'id', 'asc')
            for row in rows:
                n = n + 1
                last_id = ObjectId(row['id'])
                app_id = row.get('app_id')
                if not app_id:
                    continue

                if app_id not in self.signatures:
                    self.signatures[app_id] = {}
                signatures = self.signatures[app_id]
                signatures['signature_version'] = row.get('signature_version', '1.0')
                signatures['signature_field'] = row.get('signature_field', 'sign')
                signatures['signature_position'] = row.get('signature_position', 'body')
                signatures['signature_key'] = row.get('signature_key', '')
                signatures['expired_time'] = row.get('expired_time', 0)
            offset += n
        LOG.info('loading app_config by %s finished.', filters['type'])
        
    @tornado.gen.coroutine
    def reload_applications_from_mongo(self, application_model):
        db_proxy = DbProxy()
        limit = 100
        offset = 0
        last_id = None
        n = limit
        filters = {'type': '1910', 'obsoleted': False}
        while n >= limit:
            n = 0
            if last_id:
                filters['id__gt'] = last_id
            LOG.info('loading applications by %s from db offset:%d', filters['type'], offset)
            rows, total = yield db_proxy.query_list_mongo(application_model, filters, limit, 0, 'id', 'asc')
            for row in rows:
                n = n + 1
                last_id = ObjectId(row['id'])
                app_id = row.get('id')

                if app_id not in self.applications:
                    self.applications[app_id] = {}
                app_info = self.applications[app_id]
                app_info['name'] = row.get('name')
                app_info['type'] = row.get('type')
                app_info['code'] = row.get('code')
                app_info['cooperation_mode'] = row.get('cooperation_mode')
                app_info['expire_time'] = row.get('expire_time', 0)
                app_info['app_key'] = row.get('appKey')
        LOG.info('loading applications by %s finished.', filters['type'])
    
    def get_application_info(self, app_id):
        return self.applications.get(app_id, {})
    
    def get_signature_info(self, app_id):
        if app_id in self.signatures:
            return self.signatures[app_id]
        elif app_id in self.applications:
            return {
                'signature_version': '1.0',
                'signature_field': 'sign',
                'signature_position': 'body',
                'signature_key': self.applications[app_id]['app_key']
            }
        return False
        
ac_manager = AccessControlManager()

def whiteip_access_limit(request):
    app_id = request.headers.get('ApiAccessKey')
    if not app_id:
        LOG.warning('handled a request:%s and verifing whiteip access limit while api_access_key not given.', request.path)
        return False, 'Unknown access part'
    if app_id not in ac_manager.white_list:
        LOG.warning('handled a request:%s and verifing whiteip access limit while white list by app_id:%s not found.', request.path, str(app_id))
        return False, 'Access denied.'
    remote_ip = request.remote_ip
    if remote_ip not in ac_manager.white_list[app_id]:
        LOG.warning('handled a request:%s and verifing whiteip access limit while remote_ip:%s by app_id:%s not in white list.', request.path, str(remote_ip), str(app_id))
        return False, 'Access denied.'
    if app_id in ac_manager.black_list and remote_ip in ac_manager.black_list[app_id]:
        LOG.warning('handled a request:%s and verifing whiteip access limit while remote_ip:%s by app_id:%s in black list.', request.path, str(remote_ip), str(app_id))
        return False, 'Access denied.'
    return True, 'Success'

def signature_access_limit(request):
    app_id = request.headers.get('ApiAccessKey')
    if not app_id:
        LOG.warning('handled a request:%s and verifing signature access limit while api_access_key not given.', request.path)
        return False, 'Unknown access part'

    signature_info = ac_manager.get_signature_info(app_id)
    if not signature_info:
        LOG.warning('handled a request:%s and verifing signature access limit while signature by app_id:%s not found.', request.path, str(app_id))
        return False, 'Invalid api access key.'

    if signature_info.get('expired_time'):
        if time.time() > tofloat(signature_info.get('expired_time')):
            LOG.warning('handled a request:%s and verifing signature access limit while signature by app_id:%s not expired.', request.path, str(app_id))
            return False, 'Api expired.'

    params = None

    signature_version = signature_info.get('signature_version')
    signature_field = signature_info.get('signature_field')
    if signature_info.get('signature_position') == 'header':
        signature_value = request.headers.get(signature_field)
    else:
        if request.method == 'GET':
            params = args_as_dict(request)
        else:
            params = request_body_as_json(request)
        signature_value = params.get(signature_field)
    
    if not signature_value:
        LOG.warning('handled a request:%s and verifing signature access limit by app_id:%s while signature on field:%s in %s not filled.', request.path, str(app_id), signature_field, signature_info.get('signature_position'))
        return False, 'No signature part'
    if params is None:
        if request.method == 'GET':
            params = args_as_dict(request)
        else:
            params = request_body_as_json(request)
    if signature_version == '1.0':
        # md5
        verify_result = verify_inputs_data_sign_md5(signature_info.get('signature_key'), params, sign_field=signature_field, sign_value=signature_value)
        if not verify_result:
            LOG.warning('handled a request:%s and verifing signature access limit by app_id:%s while signature:%s on field:%s by md5 was invalid.', request.path, str(app_id), signature_value, signature_field)
            return False, 'Invalid signature part'
    elif signature_version == '2.0':
        # hmac
        verify_result = verify_inputs_data_sign_hmac(signature_info.get('signature_key'), params, sign_field=signature_field, sign_value=signature_value)
        if not verify_result:
            LOG.warning('handled a request:%s and verifing signature access limit by app_id:%s while signature:%s on field:%s by hmac was invalid.', request.path, str(app_id), signature_value, signature_field)
            return False, 'Invalid signature part'
    else:
        LOG.warning('handled a request:%s and verifing signature access limit by app_id:%s while signature by version:%s not supported.', request.path, str(app_id), signature_version)
        return False, 'Unknown configured signature version'
    
    return True, 'Success'

default_access_controls = [whiteip_access_limit, signature_access_limit]

def public_fields_control(request):
    params = None
    if params is None:
        if request.method == 'GET':
            params = args_as_dict(request)
        else:
            params = request_body_as_json(request)
    must_list = ["payload", "requestSn"]
    for ml in must_list:
        if not params.get(ml):
            return False, "lack %s." % ml
    return True, "success."

paltformaddition_access_controls = [public_fields_control]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import urllib
import re
import http
import time
import traceback
import tornado.httpclient
import tornado.gen
import tornado.escape
from tornado.ioloop import IOLoop
from concurrent.futures import ThreadPoolExecutor
from tornado.concurrent import Future, run_on_executor
import asyncio
import zeep
from .tornadozeep import TornadoAsyncTransport
from .exceptionreporter import ExceptionReporter

tornado.httpclient.AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

logging.getLogger('tornado.curl_httpclient').setLevel(logging.WARN)
logging.getLogger('zeep.wsdl.wsdl').setLevel(logging.INFO)
logging.getLogger('zeep.xsd.schema').setLevel(logging.INFO)
logging.getLogger('zeep.transports').setLevel(logging.INFO)

LOG = logging.getLogger('async.request')

@tornado.gen.coroutine
def async_get(url, params=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('GET', url, params, body=None, json=None, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body.decode()

@tornado.gen.coroutine
def async_get_bytes(url, params=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('GET', url, params, body=None, json=None, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body

@tornado.gen.coroutine
def async_post(url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('POST', url, params, body=body, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body.decode()

@tornado.gen.coroutine
def async_post_bytes(url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    response = yield async_http_request('POST', url, params, body=body, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    return response.code, response.body

@tornado.gen.coroutine
def async_post_json(url, params=None, json=None, headers=None, verify_cert=True, **kwargs):
    response = None
    try:
        response = yield async_http_request('POST', url, params, body=None, json=json, headers=headers, verify_cert=verify_cert, **kwargs)
    except Exception as e:
        LOG.error("query %s with params:%s and json:%s failed with error:%s", url, str(params), str(json), str(e))
        return False, str(e)
    if response.code == http.HTTPStatus.OK:
        try:
            data = tornado.escape.json_decode(response.body.decode())
            return True, data
        except Exception as e:
            LOG.error("query %s with params:%s and json:%s while decode response data:%s failed with error:%s", url, str(params), str(json), str(response.body.decode()), str(e))
            return False, str(e) + ' ' + response.body.decode()
    res = str(response.reason) + ' ' + response.body.decode()
    return False, res

@tornado.gen.coroutine
def async_http_request(method, url, params=None, body='', json=None, headers=None, verify_cert=True, **kwargs):
    proxies = kwargs.pop('proxies', None)
    if params:
        qrs = [k + '=' + urllib.parse.quote(v) for k,v in params.items() ]
        sep = '&' if '?' in url else '?'
        url += sep + '&'.join(qrs)
    if json:
        body = tornado.escape.json_encode(json)
        if not headers:
            headers = {}
        headers['Content-Type'] = 'application/json'
    
    request = tornado.httpclient.HTTPRequest(url, method=method, body=body, headers=headers, validate_cert=verify_cert, **kwargs)
    client = tornado.httpclient.AsyncHTTPClient()
    prepare_request_proxies(request, proxies)
    
    response = yield client.fetch(request, raise_error=False)
    if response.code != http.HTTPStatus.OK:
        ExceptionReporter().report(key='HTTP-'+str(response.code), typ='HTTPQuery', 
            endpoint=url,
            method=method,
            inputs=body if body else tornado.escape.json_encode(params) if params else '',
            outputs=str(response.body.decode()),
            content=str(response.body.decode()),
            level='ERROR'
        )
    return response

def parse_http_proxies(proxy_url):
    '''parse http proxies from proxy url
    parse http://username:password@host:port as dict of schema://username:password@host:port
    '''
    if not proxy_url:
        return {}
    founds = re.search(r'(?P<schema>https?)://(?:(?P<username>[A-Za-z0-9\%\_\-\+\=]*)(?:\:(?P<password>[A-Za-z0-9\%\_\-\+\=]*))?\@)?(?P<host>(?:\w+)(?:\.\w+)*)(?:\:(?P<port>\d+))?', proxy_url)
    if not founds:
        return {}
    return {
        'schema': founds['schema'],
        'host': founds['host'],
        'port': int(founds['port'] if founds['port'] else 80),
        'username': urllib.parse.unquote(founds['username']) if founds['username'] else '',
        'password': urllib.parse.unquote(founds['password']) if founds['password'] else ''
    }

def prepare_request_proxies(request, proxies):
    if proxies:
        if 'host' in proxies and 'port' in proxies:
            request.proxy_host = proxies.get('host')
            request.proxy_port = int(proxies.get('port'))
        else:
            schema = 'http'
            if request.url.startswith('https'):
                schema = 'https'
            if schema in proxies:
                proxy_info = parse_http_proxies(proxies.get(schema))
                if proxy_info and proxy_info.get('host', False):
                    request.proxy_host = proxy_info.get('host', '')
                    request.proxy_port = int(proxy_info.get('port', 80))
                    if proxy_info.get('username', False):
                        request.proxy_username = proxy_info['username']
                    if proxy_info.get('password', False):
                        request.proxy_password = proxy_info['password']
        
# class AsyncSoapExecutor():
#     """
#     """
#     def __init__(self, ioloop = None):
#         # self.executor = ThreadPoolExecutor(max_workers=50)
#         # self.io_loop = ioloop or asyncio.new_event_loop()
#         pass

#     @tornado.gen.coroutine
#     def soap_request(self, future, endpoint, action, params, **kwargs):
#         proxies = kwargs.pop('proxies', None)
#         transport = kwargs.pop('transport', None)
#         uuid = kwargs.pop('uuid', None)
#         err = ''
#         context = ''
#         try:
#             web_client = zeep.AsyncClient(endpoint, transport=transport)
#             if not hasattr(web_client.service, action):
#                 return None, 'There is no action %s in wsdl %s' % (action, endpoint)
#             soap_method = web_client.service[action]
#             if isinstance(params, dict):
#                 context = yield soap_method(**params)
#             elif isinstance(params, list):
#                 context = yield soap_method(*params)
#             else:
#                 context = yield soap_method(params)
#         except Exception as e:
#             LOG.error(traceback.format_exc())
#             err = str(e)
#             ExceptionReporter().report(key='SOAP-'+str('failed'), typ='SOAPQuery', 
#                 endpoint=endpoint,
#                 method=str(action),
#                 inputs=str(params),
#                 outputs=str(context),
#                 content=err,
#                 level='ERROR'
#             )
#             # context = '{"code": -1, "message": %s}' % str(e)
#         if future:
#             future.set_result(context)
#         # print('result', context)
#         LOG.debug('%s %s %d response:\n%s', endpoint, action, uuid, context)
#         return context, err

class _wsdl_content(str):
    def __init__(self, endpoint: str) -> None:
        super().__init__()
        self.endpoint = endpoint
        self.content = bytes()
        self.expires = 0

    def read(self) -> bytes:
        return self.content

    def set_content(self, content: bytes) -> None:
        self.content = content
        self.expires = time.time() + 300

    def __str__(self) -> str:
        return self.endpoint

_wsdl_caches = {}

@tornado.gen.coroutine
def async_soap_request(endpoint, action, params, **kwargs):
    proxies = kwargs.pop('proxies', None)
    transport = kwargs.pop('transport', None)
    uuid = kwargs.pop('uuid', None)
    err = ''
    context = ''
    try:
        t1 = time.time()
        wsdl = None
        if endpoint in _wsdl_caches:
            wsdl = _wsdl_caches[endpoint]
            if wsdl.expires < t1:
                wsdl = None
        if wsdl is None:
            wsdl_code, wsdl_content = yield async_get_bytes(endpoint)
            if wsdl_code != 200:
                if wsdl_code == 301:
                    pass
                return None, 'Query wsdl %s failed with error:%s' % (endpoint, str(wsdl_content))
            wsdl = _wsdl_content(endpoint)
            wsdl.set_content(wsdl_content)
            _wsdl_caches[endpoint] = wsdl
        
        web_client = zeep.AsyncClient(wsdl, transport=transport)
        if not hasattr(web_client.service, action):
            return None, 'There is no action %s in wsdl %s' % (action, endpoint)
        soap_method = web_client.service[action]
        if isinstance(params, dict):
            context = yield soap_method(**params)
        elif isinstance(params, list):
            context = yield soap_method(*params)
        else:
            context = yield soap_method(params)
        ti = time.time() - t1
        if ti > 2:
            LOG.warning('query wsdl %s action %s cost too much time (%.2fs)', endpoint, action, ti)
    except Exception as e:
        LOG.error(traceback.format_exc())
        err = str(e)
        ExceptionReporter().report(key='SOAP-'+str('failed'), typ='SOAPQuery', 
            endpoint=endpoint,
            method=str(action),
            inputs=str(params),
            outputs=str(context),
            content=err,
            level='ERROR'
        )
        # context = '{"code": -1, "message": %s}' % str(e)
    # print('result', context)
    LOG.debug('%s %s response:\n%s', endpoint, action, context)
    return context, err

def main():
    import tornado.web
    from tornado.ioloop import IOLoop

    routes = [
    ]

    class GeneralTornadoHandler(tornado.web.RequestHandler):
        """
        """

        def initialize(self, callback, methods):
            self.callbacks = {}
            if isinstance(methods, str):
                methods = [methods]
            for method in methods:
                method = method.upper()
                self.callbacks[method] = callback

        @tornado.gen.coroutine
        def get(self):
            yield self._do_callback(self.callbacks.get('GET'))

        @tornado.gen.coroutine
        def post(self):
            yield self._do_callback(self.callbacks.get('POST'))
        
        @tornado.gen.coroutine
        def _do_callback(self, cb):
            if callable(cb):
                response = yield cb(self, self.request)
                if isinstance(response, str):
                    self.write(response)
                    self.finish()
                else:
                    print("====== response", response)
                    pass
            else:
                print('not callable cb:', cb)

    def tornado_route(rule, **options):
        def decorator(f):
            # endpoint = options.pop('endpoint', None)
            # if not endpoint:
            #     endpoint = f.__name__

            routes.append((rule, GeneralTornadoHandler, dict(callback=f, methods=options.pop('methods', ['GET']))))
            return f
        return decorator

    app2 = tornado.web.Application(routes)
    app2.listen(8021)
    
    IOLoop.instance().start()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import weakref
from abc import ABC, abstractmethod
from collections.abc import Coroutine

from represent import ReprHelper
from sqlalchemy.engine import Engine
from sqlalchemy.exc import StatementError
from sqlalchemy import util
from sqlalchemy.log import Identified
from sqlalchemy.util.concurrency import greenlet_spawn

from .exc import AlreadyQuit, BlockingWarning


class AsyncEngine(Identified, ABC):
    def __init__(self, engine: Engine, **kwargs):
        self._engine = engine

        self._engine_worker = None

    @abstractmethod
    def _make_worker(self, *, branch_from=None):
        raise NotImplementedError

    async def _run_in_thread(_self, _func, *args, **kwargs):
        """Unlike the public-facing `run_in_thread` method, we want this one
        to let us call SQLAlchemy methods like normal internally.
        """
        if _self._engine_worker is None:
            _self._engine_worker = _self._make_worker()

        return await _self._engine_worker.run(_func, args, kwargs)

    async def run_in_thread(self, func, *args):
        """Run a synchronous function in the engine's worker thread.
        Example:
            The following blocking function:
            .. code-block:: python
                some_fn(engine.sync_engine)
            can be called like this instead:
            .. code-block:: python
                await engine.run_in_thread(some_fn, engine.sync_engine)
        Parameters:
            func: A synchronous function.
            args: Positional arguments to be passed to `func`. If you need to
                pass keyword arguments, then use :func:`functools.partial`.
        """
        if self._engine_worker is None:
            self._engine_worker = self._make_worker()

        return await self._engine_worker.run(func, args)

    @property
    def dialect(self):
        return self._engine.dialect

    @property
    def _has_events(self):
        return self._engine._has_events

    @property
    def logger(self):
        return self._engine.logger

    @property
    def _execution_options(self):
        return self._engine._execution_options

    @property
    def sync_engine(self):
        """Public property of the underlying SQLAlchemy engine."""
        return self._engine

    def connect(self):
        """Like :meth:`Engine.connect <sqlalchemy.engine.Engine.connect>`, but
        returns an awaitable that can also be used as an asynchronous context
        manager.
        Examples:
            .. code-block:: python
                conn = await engine.connect()
                await conn.execute(...)
                await conn.close()
            .. code-block:: python
                async with engine.connect() as conn:
                    await conn.execute(...)
        """
        return _ConnectionContextManager(self._make_async_connection())

    async def _make_async_connection(self):
        worker = self._make_worker()
        try:
            connection = await worker.run(self._engine.connect)
        except Exception:
            await worker.quit()
            raise
        return AsyncConnection(connection, worker, self)

    def begin(self, close_with_result=False):
        """Like :meth:`Engine.begin <sqlalchemy.engine.Engine.begin>`, but
        returns an asynchronous context manager.
        Example:
            .. code-block:: python
                async with engine.begin():
                    await engine.execute(...)
        """
        return _EngineTransactionContextManager(self, close_with_result)

    async def execute(self, *args, **kwargs):
        """Like :meth:`Engine.execute <sqlalchemy.engine.Engine.execute>`, but
        is a coroutine that returns an :class:`AsyncioResultProxy`.
        Example:
            .. code-block:: python
                result = await engine.execute(...)
                data = await result.fetchall()
        .. warning::
            Make sure to explicitly call :meth:`AsyncioResultProxy.close` if the
            :class:`~sqlalchemy.engine.ResultProxy` has pending rows remaining
            otherwise it will be closed during garbage collection. With SQLite,
            this will raise an exception since the DBAPI connection was created
            in a different thread.
        """
        rp = await self._run_in_thread(self._engine.execute, *args, **kwargs)
        return AsyncResultProxy(rp, self._run_in_thread)

    async def scalar(self, *args, **kwargs):
        """Like :meth:`Connection.scalar <sqlalchemy.engine.Engine.scalar>`,
        but is a coroutine.
        """
        rp = await self.execute(*args, **kwargs)
        return await rp.scalar()

    async def has_table(self, table_name, schema=None):
        """Like :meth:`Engine.has_table <sqlalchemy.engine.Engine.has_table>`,
        but is a coroutine.
        """
        return await self._run_in_thread(
            self._engine.has_table, table_name, schema)

    async def table_names(
            self, schema=None, connection: 'AsyncConnection' = None):
        """Like :meth:`Engine.table_names <sqlalchemy.engine.Engine.table_names>`,
        but is a coroutine.
        """
        run_in_thread = self._run_in_thread

        if connection is not None:
            run_in_thread = connection._run_in_thread
            connection = connection._connection

        return await run_in_thread(self._engine.table_names, schema, connection)

    def run_callable(self, callable_, *args, **kwargs):
        """Like :meth:`Engine.run_callable\
        <sqlalchemy.engine.Engine.run_callable>`.
        .. warning::
            This method blocks. It exists so that we can warn the user if
            they try to use an async engine for table reflection:
            .. code-block:: python
                Table(..., autoload_with=engine)
        """
        warnings.warn(
            'The AsyncEngine has been called in a blocking fashion, e.g. with '
            'Table(..., autoload_with=engine). You may wish to run it in a '
            'separate thread to avoid blocking the event loop. You can use '
            'Table(..., autoload_with=engine.sync_engine) to opt out of the '
            'warning for this blocking behaviour.',
            BlockingWarning)
        self._engine.run_callable(callable_, *args, **kwargs)

    def __repr__(self):
        r = ReprHelper(self)
        r.parantheses = ('<', '>')
        r.positional_from_attr('_engine')
        return str(r)

    def __getattr__(self, item):
        msg = '{!r} object has no attribute {!r}.'.format(
            self.__class__.__name__, item)

        if item == '_run_visitor':
            raise AttributeError(
                msg + ' Did you try to use Table.create(engine) or similar? '
                'You must use Table.create(engine.sync_engine) instead, which'
                'is a blocking function. Consider using sqlalchemy.schema.'
                'CreateTable instead'
            )
        elif item == 'dispatch':
            raise AttributeError(
                msg + ' Did you try to use event.listen(engine, ...)? You must '
                'use event.listen(engine.sync_engine, ...) instead.'
            )

        raise AttributeError(msg)

    async def raw_connection(self):
        """Return a "raw" DBAPI connection from the connection pool.

        .. seealso::

            :ref:`dbapi_connections`

        """
        return await greenlet_spawn(self.sync_engine.raw_connection)

class AsyncConnection:
    """Mostly like :class:`sqlalchemy.engine.Connection` except some of the
    methods are coroutines.
    """
    def __init__(self, connection, worker, engine):
        self._connection = connection
        self._worker = worker
        self._engine_ref = weakref.ref(engine)

    async def _run_in_thread(_self, _func, *args, **kwargs):
        return await _self._worker.run(_func, args, kwargs)

    async def run_in_thread(self, func, *args):
        """Run a synchronous function in the connection's worker thread.
        Example:
            The following blocking function:
            .. code-block:: python
                some_fn(conn.sync_connection)
            can be called like this instead:
            .. code-block:: python
                await engine.run_in_thread(some_fn, conn.sync_connection)
        Parameters:
            func: A synchronous function.
            args: Positional arguments to be passed to `func`. If you need to
                pass keyword arguments, then use :func:`functools.partial`.
        """
        return await self._worker.run(func, args)

    @property
    def _engine(self):
        return self._engine_ref()

    @property
    def sync_connection(self):
        """Public property of the underlying SQLAlchemy connection."""
        return self._connection

    @property
    def dialect(self):
        return self._connection.dialect

    async def execute(self, *args, **kwargs):
        """Like :meth:`Connection.execute <sqlalchemy.engine.Connection.execute>`,
        but is a coroutine that returns an :class:`AsyncioResultProxy`.
        Example:
            .. code-block:: python
                result = await conn.execute(...)
                data = await result.fetchall()
        .. warning::
            Make sure to explicitly call :meth:`AsyncioResultProxy.close` if the
            :class:`~sqlalchemy.engine.ResultProxy` has pending rows remaining
            otherwise it will be closed during garbage collection. With SQLite,
            this will raise an exception since the DBAPI connection was created
            in a different thread.
        """
        try:
            rp = await self._run_in_thread(
                self._connection.execute, *args, **kwargs)
        except AlreadyQuit:
            raise StatementError("This Connection is closed.", None, None, None)

        return AsyncResultProxy(rp, self._run_in_thread)

    def connect(self):
        """Like :meth:`Connection.connect <sqlalchemy.engine.Connection.connect>`,
        but is a coroutine.
        """
        return _ConnectionContextManager(self._make_async_connection())

    async def _make_async_connection(self):
        worker = self._engine._make_worker(branch_from=self._worker)
        try:
            connection = await worker.run(self._connection.connect)
        except Exception:
            await worker.quit()
            raise
        return AsyncConnection(connection, worker, self._engine)

    async def scalar(self, *args, **kwargs):
        """Like :meth:`Connection.scalar <sqlalchemy.engine.Connection.scalar>`,
        but is a coroutine.
        """
        rp = await self.execute(*args, **kwargs)
        return await rp.scalar()

    async def close(self, *args, **kwargs):
        """Like :meth:`Connection.close <sqlalchemy.engine.Connection.close>`,
        but is a coroutine.
        """
        try:
            res = await self._run_in_thread(
                self._connection.close, *args, **kwargs)
            await self._worker.quit()
        except AlreadyQuit:
            raise StatementError("This Connection is closed.", None, None, None)

        return res

    @property
    def closed(self):
        """Like the :attr:`Connection.closed\
        <sqlalchemy.engine.Connection.closed>` attribute.
        """
        return self._connection.closed

    def begin(self):
        """Like :meth:`Connection.begin <sqlalchemy.engine.Connection.begin>`,
        but returns an awaitable that can also be used as an asynchronous
        context manager.
        Examples:
            .. code-block:: python
                async with conn.begin() as trans:
                    await conn.execute(...)
                    await conn.execute(...)
            .. code-block:: python
                trans = await conn.begin():
                await conn.execute(...)
                await conn.execute(...)
                await trans.commit()
        """
        return _TransactionContextManager(self._begin())

    async def _begin(self):
        try:
            transaction = await self._run_in_thread(self._connection.begin)
        except AlreadyQuit:
            raise StatementError("This Connection is closed.", None, None, None)

        return AsyncTransaction(transaction, self._run_in_thread)

    def begin_nested(self):
        """Like :meth:`Connection.begin_nested\
        <sqlalchemy.engine.Connection.begin_nested>`, but returns an awaitable
        that can also be used as an asynchronous context manager.
        .. seealso:: :meth:`begin` for examples.
        """
        return _TransactionContextManager(self._begin_nested())

    async def _begin_nested(self):
        try:
            transaction = await self._run_in_thread(self._connection.begin_nested)
        except AlreadyQuit:
            raise StatementError("This Connection is closed.", None, None, None)

        return AsyncTransaction(transaction, self._run_in_thread)

    def in_transaction(self):
        """Like :meth:`Connection.in_transaction\
        <sqlalchemy.engine.Connection.in_transaction>`.
        """
        return self._connection.in_transaction()

    def run_callable(self, callable_, *args, **kwargs):
        """Like :meth:`Connection.run_callable\
        <sqlalchemy.engine.Connection.run_callable>`.
        .. warning::
            This method blocks. It exists so that we can warn the user if
            they try to use an async connection for table reflection:
            .. code-block:: python
                Table(..., autoload_with=connection)
        """
        warnings.warn(
            'The AsyncConnection has been called in a blocking fashion, e.g. '
            'with Table(..., autoload_with=connection). You may wish to run it '
            'in a separate thread to avoid blocking the event loop. You can '
            'use Table(..., autoload_with=connection.sync_connection) to opt '
            'out of the warning for this blocking behaviour.',
            BlockingWarning)
        self._connection.run_callable(callable_, *args, **kwargs)

    def __getattr__(self, item):
        msg = '{!r} object has no attribute {!r}.'.format(
            self.__class__.__name__, item)

        if item == '_run_visitor':
            raise AttributeError(
                msg + ' Did you try to use Table.create(connection) or '
                'similar? You must use Table.create(connection.sync_connection'
                ') instead, which is a blocking function. Consider using '
                'sqlalchemy.schema.CreateTable instead.'
            )
        elif item == 'dispatch':
            raise AttributeError(
                msg + ' Did you try to use event.listen(connection, ...)? You'
                'must use event.listen(connection.sync_connection, ...) instead.'
            )

        raise AttributeError(msg)

    async def run_sync(self, fn, *arg, **kw):
        """Invoke the given sync callable passing self as the first argument.

        This method maintains the asyncio event loop all the way through
        to the database connection by running the given callable in a
        specially instrumented greenlet.

        E.g.::

            with async_engine.begin() as conn:
                await conn.run_sync(metadata.create_all)

        .. note::

            The provided callable is invoked inline within the asyncio event
            loop, and will block on traditional IO calls.  IO within this
            callable should only call into SQLAlchemy's asyncio database
            APIs which will be properly adapted to the greenlet context.

        .. seealso::

            :ref:`session_run_sync`
        """

        conn = self.sync_connection

        return await greenlet_spawn(fn, conn, *arg, **kw)


class AsyncTransaction:
    """Mostly like :class:`sqlalchemy.engine.Transaction` except some of the
    methods are coroutines.
    """
    def __init__(self, transaction, run_in_thread):
        self._transaction = transaction
        self._run_in_thread = run_in_thread

    async def commit(self):
        """Like :meth:`Transaction.commit <sqlalchemy.engine.Transaction.commit>`,
        but is a coroutine.
        """
        return await self._run_in_thread(self._transaction.commit)

    async def rollback(self):
        """Like :meth:`Transaction.rollback <sqlalchemy.engine.Transaction.rollback>`,
        but is a coroutine.
        """
        return await self._run_in_thread(self._transaction.rollback)

    async def close(self):
        """Like :meth:`Transaction.close <sqlalchemy.engine.Transaction.close>`,
        but is a coroutine.
        """
        return await self._run_in_thread(self._transaction.close)


class _AsyncResultProxyIterator:
    def __init__(self, result_proxy, run_in_thread):
        self._result_proxy = result_proxy
        self._run_in_thread = run_in_thread

    def __aiter__(self):
        return self

    async def __anext__(self):
        row = await self._run_in_thread(self._result_proxy.fetchone)
        if row is None:
            raise StopAsyncIteration()
        else:
            return row


class AsyncResultProxy:
    """Mostly like :class:`sqlalchemy.engine.ResultProxy` except some of the
    methods are coroutines.
    """
    def __init__(self, result_proxy, run_in_thread):
        self._result_proxy = result_proxy
        self._run_in_thread = run_in_thread

    def __aiter__(self):
        return _AsyncResultProxyIterator(
            self._result_proxy,
            self._run_in_thread)

    async def fetchone(self):
        """Like :meth:`ResultProxy.fetchone\
        <sqlalchemy.engine.ResultProxy.fetchone>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.fetchone)

    async def fetchmany(self, size=None):
        """Like :meth:`ResultProxy.fetchmany\
        <sqlalchemy.engine.ResultProxy.fetchmany>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.fetchmany, size=size)

    async def fetchall(self):
        """Like :meth:`ResultProxy.fetchall\
        <sqlalchemy.engine.ResultProxy.fetchall>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.fetchall)

    async def scalar(self):
        """Like :meth:`ResultProxy.scalar\
        <sqlalchemy.engine.ResultProxy.scalar>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.scalar)

    async def first(self):
        """Like :meth:`ResultProxy.first\
        <sqlalchemy.engine.ResultProxy.first>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.first)

    async def keys(self):
        """Like :meth:`ResultProxy.keys\
        <sqlalchemy.engine.ResultProxy.keys>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.keys)

    async def close(self):
        """Like :meth:`ResultProxy.close\
        <sqlalchemy.engine.ResultProxy.close>`, but is a coroutine.
        """
        return await self._run_in_thread(self._result_proxy.close)

    @property
    def returns_rows(self):
        """Like the :attr:`ResultProxy.returns_rows\
        <sqlalchemy.engine.ResultProxy.returns_rows>` attribute.
        """
        return self._result_proxy.returns_rows

    @property
    def rowcount(self):
        """Like the :attr:`ResultProxy.rowcount\
        <sqlalchemy.engine.ResultProxy.rowcount>` attribute.
        """
        return self._result_proxy.rowcount

    @property
    def inserted_primary_key(self):
        """Like the :attr:`ResultProxy.inserted_primary_key\
        <sqlalchemy.engine.ResultProxy.inserted_primary_key>` attribute.
        """
        return self._result_proxy.inserted_primary_key


class _BaseContextManager(Coroutine):
    """Allow ``async with <coroutine>`` or ``await <coroutine>``."""
    __slots__ = ('_coro', '_result')

    def __init__(self, coro):
        self._coro = coro
        self._result = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._result = await self._coro
        return self._result

    def __enter__(self):
        raise TypeError(
            'Use async with instead. This error can occur when trying to '
            'pass the AsyncEngine to something that expects a normal engine, '
            'e.g. MetaData.reflect(). You can use engine.sync_engine, but be '
            'aware that the function will block. You should run it in a '
            'separate thread. See also connection.sync_connection'
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class _ConnectionContextManager(_BaseContextManager):
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._result.close()


class _TransactionContextManager(_BaseContextManager):
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and self._result._transaction.is_active:
            try:
                await self._result.commit()
            except:
                with util.safe_reraise():
                    await self._result.rollback()
        else:
            await self._result.rollback()


class _EngineTransactionContextManager:
    __slots__ = ('_engine', '_close_with_result', '_context', '_worker')

    def __init__(self, engine: AsyncEngine, close_with_result):
        self._engine = engine
        self._close_with_result = close_with_result
        self._worker = self._engine._make_worker()

    async def _run_in_thread(_self, _func, *args, **kwargs):
        return await _self._worker.run(_func, args, kwargs)

    async def __aenter__(self):
        self._context = await self._run_in_thread(
            self._engine._engine.begin, self._close_with_result)

        conn = await self._run_in_thread(self._context.__enter__)
        return AsyncConnection(conn, self._worker, self._engine)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._run_in_thread(
            self._context.__exit__, exc_type, exc_val, exc_tb)


class ThreadWorker(ABC):
    @abstractmethod
    async def run(self, func, args=(), kwargs=None):
        raise NotImplementedError

    @abstractmethod
    async def quit(self):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lycium.kafka.kafkaWorker import KafkaWorker
from lycium.microsvc.router import Router
from lycium.microsvc.middlewares.exceptionMiddleware import ExceptionMiddleware
from lycium.microsvc.middlewares.logMiddleware import LogMiddleware
from lycium.microsvc.middlewares.websocketPackageMiddleware import WebsocketPackageMiddleware
from tornado.ioloop import IOLoop



class ApplicationService(object):
    """ 
    提供使用kafka 接收请求，并进行分发
    """
    def __init__(self,topic:str,hosts:list,app_code:str,username:str="",password:str=""):
       """ 
       :params topic 订阅的topic
       :params hosts kafka 节点的地址列表,如["127.0.0.1:9092","127.0.01:9093"]
       :params app_code 应用程序代码，用于返回错误时候区分不同的微服务
       :params username kafka 集群的用户名
       :params password kafka 集群的密码
       """
       self._topic = topic
       self._app_code = app_code
       worker = KafkaWorker(hosts=hosts,
                            private_topic=topic+"-private",
                            group_id=topic + "-group",
                            sasl_username = username,
                            sasl_password = password)
        
       self._worker = worker
       self._router = Router(self._worker)
       self.set_environment("app_code", app_code)
    
    def set_environment(self,key,value):
        """
        设置全局可用的环境变量 
        """
        self._router.set_environment(key,value)

    def add_middleware(self,middleware):
        """ 
        新增中间件
        """
        self._router.add_middleware(middleware)

    def add_read(self,module_name):
        """ 
        新增读权限的业务方法
        """
        def decorate(func):
            self._router.add_read(func, module_name)
            return func
        return decorate

    def add_create(self,module_name):
        """ 
        新增读权限的业务方法
        """
        def decorate(func):
            self._router.add_create(func, module_name)
            return func
        return decorate

    def add_update(self,module_name):
        """ 
        新增读权限的业务方法
        """
        def decorate(func):
            self._router.add_update(func, module_name)
            return func
        return decorate

    def add_delete(self,module_name):
        """ 
        新增读权限的业务方法
        """
        def decorate(func):
            self._router.add_delete(func, module_name)
            return func
        return decorate

    def init_middlewares(self):
        """ 
        加载常用的中间件
        """
        self.add_middleware(WebsocketPackageMiddleware)
        self.add_middleware(ExceptionMiddleware)
        self.add_middleware(LogMiddleware)

    def run(self):
        """ 
        服务开始运行
        """
        self.init_middlewares()
        self._worker.subscribe(self._topic, self._router.service)
        IOLoop.instance().start()